from kratos import *
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.passes.passes import lift_config_reg
from lake.modules.reg_fifo import RegFIFO
import kratos as kts


class StrgRAM(Generator):
    '''
    Storage RAM
        Does ROM/RAM from storage
    '''
    def __init__(self,
                 data_width=16,
                 banks=1,
                 memory_width=64,
                 memory_depth=512,
                 num_tiles=1,
                 rw_same_cycle=False,  # Same as separate addresses
                 read_delay=1,
                 addr_width=16,
                 prioritize_write=True):
        super().__init__("strg_ram", debug=True)

        # Generation parameters
        self.banks = banks
        self.data_width = data_width
        self.memory_width = memory_width
        self.memory_depth = memory_depth
        self.num_tiles = num_tiles
        self.rw_same_cycle = rw_same_cycle
        self.read_delay = read_delay
        self.addr_width = addr_width
        self.fw_int = int(self.memory_width / self.data_width)
        self.prioritize_write = prioritize_write
        self.bank_width = clog2(self.banks)
        self.word_width = max(1, clog2(self.fw_int))
        self.mem_addr_width = clog2(self.num_tiles * self.memory_depth)
        self.b_a_off = clog2(self.fw_int) + clog2(self.num_tiles * self.memory_depth)

        # assert banks > 1 or rw_same_cycle is True or self.fw_int > 1, \
        #     "Can't sustain throughput with this setup. Need potential bandwidth for " + \
        #     "1 write and 1 read in a cycle - try using more banks or a macro that supports 1R1W"

        # Clock and Reset
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        # Inputs + Outputs
        self._wen = self.input("wen", 1)
        self._ren = self.input("ren", 1)
        self._data_in = self.input("data_in", self.data_width)
        self._wr_addr_in = self.input("wr_addr_in", self.addr_width)
        self._rd_addr_in = self.input("rd_addr_in", self.addr_width)

        self._wr_addr = self.var("wr_addr", self.addr_width)
        self._rd_addr = self.var("rd_addr", self.addr_width)

        # Separate addressing...
        if self.rw_same_cycle:
            self.wire(self._wr_addr, self._wr_addr_in)
            self.wire(self._rd_addr, self._rd_addr_in)
        # Use the wr addr for both in this case...
        else:
            self.wire(self._wr_addr, self._wr_addr_in)
            self.wire(self._rd_addr, self._wr_addr_in)

        self._data_out = self.output("data_out", self.data_width)
        self._valid_out = self.output("valid_out", 1)

        # get relevant signals from the storage banks
        self._data_from_strg = self.input("data_from_strg", self.data_width,
                                          size=(self.banks,
                                                self.fw_int),
                                          explicit_array=True,
                                          packed=True)

        self._wen_addr = self.var("wen_addr", self.addr_width,
                                  size=self.banks,
                                  explicit_array=True,
                                  packed=True)

        self._ren_addr = self.var("ren_addr", self.addr_width,
                                  size=self.banks,
                                  explicit_array=True,
                                  packed=True)

        self._data_to_strg = self.output("data_to_strg", self.data_width,
                                         size=(self.banks,
                                               self.fw_int),
                                         explicit_array=True,
                                         packed=True)

        self._wen_to_strg = self.output("wen_to_strg", self.banks)
        self._ren_to_strg = self.output("ren_to_strg", self.banks)

        self._addr_out = self.output("addr_out", self.mem_addr_width,
                                     size=self.banks,
                                     explicit_array=True,
                                     packed=True)

        self._rd_bank = self.var("rd_bank", max(1, clog2(self.banks)))
        self.set_read_bank()
        self._rd_valid = self.var("rd_valid", 1)
        self.set_read_valid()
        if self.fw_int == 1:
            self.wire(self._valid_out, self._rd_valid)

        # Fetch width of 1 is simpler...
        if self.fw_int == 1:
            # Set data to storage
            if self.banks == 1:
                self.wire(self._wen_to_strg, self._wen)
                self.wire(self._ren_to_strg, self._ren)
                self.wire(self._data_to_strg[0], self._data_in)
                self.wire(self._addr_out[0],
                          kts.ternary(self._wen_to_strg[0],
                                      self._wr_addr[self.mem_addr_width - 1, 0],
                                      self._rd_addr[self.mem_addr_width - 1, 0]))
            else:
                for i in range(self.banks):
                    self.wire(self._data_to_strg[i], self._data_in)
                    self.add_code(self.decode_wen, idx=i)
                    self.add_code(self.decode_ren, idx=i)
                    self.wire(self._addr_out[i],
                              kts.ternary(self._wen_to_strg[i],
                                          self._wr_addr[self.mem_addr_width - 1, 0],
                                          self._rd_addr[self.mem_addr_width - 1, 0]))
            self.wire(self._data_out, self._data_from_strg[self._rd_bank])
        elif self.read_delay == 1:

            self._data_to_write = self.var("data_to_write", self.data_width)
            self._addr_to_write = self.var("addr_to_write", self.addr_width)
            self.add_code(self.set_dat_to_write)
            self.add_code(self.set_addr_to_write)

            self._write_gate = self.var("write_gate", 1)
            self._read_gate = self.var("read_gate", 1)

            self._data_combined = self.var("data_combined", self.data_width,
                                           size=self.fw_int,
                                           explicit_array=True,
                                           packed=True)

            for i in range(self.banks):
                self.wire(self._data_to_strg[i], self._data_combined)

            # read-modify-write implies we need to stall upstream
            self._ready = self.output("ready", 1)
            # Otherwise implement the state machine for read-modify-write
            self.rmw_fsm = self.add_fsm("r_w_seq", reset_high=False)
            IDLE = self.rmw_fsm.add_state("IDLE")
            READ = self.rmw_fsm.add_state("READ")
            MODIFY = self.rmw_fsm.add_state("MODIFY")
            DEFAULT = self.rmw_fsm.add_state("_DEFAULT")
            self.rmw_fsm.output(self._ready)
            self.rmw_fsm.output(self._valid_out)
            self.rmw_fsm.output(self._data_out)
            self.rmw_fsm.output(self._write_gate)
            self.rmw_fsm.output(self._read_gate)

            # In IDLE we go to a read state if reading, and modify state
            # if writing....
            IDLE.next(IDLE, ~(self._wen) & ~(self._ren))
            IDLE.next(READ, self._ren & ~self._wen)
            IDLE.next(MODIFY, self._wen)
            # OUT
            IDLE.output(self._ready, 1)
            IDLE.output(self._valid_out, 0)
            IDLE.output(self._data_out, 0)
            IDLE.output(self._write_gate, 0)
            IDLE.output(self._read_gate, 1)

            # In READ, we effectively use the same transitions as IDLE
            READ.next(IDLE, ~self._wen & ~self._ren)
            READ.next(READ, self._ren & ~self._wen)
            READ.next(MODIFY, self._wen)
            # OUT
            READ.output(self._ready, 1)
            READ.output(self._valid_out, 1)
            READ.output(self._data_out,
                        self._data_from_strg[self._rd_bank][self._addr_to_write[self.word_width - 1, 0]])
            READ.output(self._write_gate, 0)
            READ.output(self._read_gate, 1)

            # In MODIFY we always go back to idle
            MODIFY.next(IDLE, const(1, 1))
            MODIFY.output(self._ready, 0)
            MODIFY.output(self._valid_out, 0)
            MODIFY.output(self._data_out, 0)
            MODIFY.output(self._write_gate, 1)
            MODIFY.output(self._read_gate, 0)

            # In DEFAULT we always stick in DEFAULT because it's over...
            DEFAULT.next(DEFAULT, const(1, 1))
            DEFAULT.output(self._ready, 0)
            DEFAULT.output(self._valid_out, 0)
            DEFAULT.output(self._data_out, 0)
            DEFAULT.output(self._write_gate, 0)
            DEFAULT.output(self._read_gate, 0)

            self.rmw_fsm.set_start_state(IDLE)

            if self.banks == 1:
                self.wire(self._ren_to_strg, (self._wen | self._ren) & self._read_gate)
                self.wire(self._wen_to_strg, self._write_gate)
            else:
                for i in range(self.banks):
                    self.add_code(self.set_wen_rmw, idx=i)
                    self.add_code(self.set_ren_rmw, idx=i)

            for i in range(self.banks):
                self.add_code(self.set_addr_rmw, idx=i)
            for i in range(self.fw_int):
                self.add_code(self.set_data_combined, idx=i)
        # If read delay is 0, we can rmw in the same cycle (TIMING?)

    def set_read_bank(self):
        if self.banks == 1:
            self.wire(self._rd_bank, kts.const(0, 1))
        else:
            # The read bank is comb if no delay, otherwise delayed
            if self.read_delay == 1:
                @always_ff((posedge, "clk"), (negedge, "rst_n"))
                def read_bank_ff(self):
                    if ~self._rst_n:
                        self._rd_bank = 0
                    else:
                        self._rd_bank = \
                            self._rd_addr[self.b_a_off + self.bank_width - 1, self.b_a_off]
                self.add_code(read_bank_ff)
            else:
                @always_comb
                def read_bank_comb(self):
                    self._rd_bank = \
                        self._rd_addr[self.b_a_off + self.bank_width - 1, self.b_a_off]
                self.add_code(read_bank_comb)

    def set_read_valid(self):
        # The read bank is comb if no delay, otherwise delayed
        if self.read_delay == 1:
            if self.rw_same_cycle:
                @always_ff((posedge, "clk"), (negedge, "rst_n"))
                def read_valid_ff(self):
                    if ~self._rst_n:
                        self._rd_valid = 0
                    else:
                        # Don't need write priority if both go at once
                        self._rd_valid = self._ren
                self.add_code(read_valid_ff)
            else:
                @always_ff((posedge, "clk"), (negedge, "rst_n"))
                def read_valid_ff(self):
                    if ~self._rst_n:
                        self._rd_valid = 0
                    else:
                        # Assumes write priority
                        self._rd_valid = self._ren & ~self._wen
                self.add_code(read_valid_ff)
        else:
            if self.rw_same_cycle:
                self.wire(self._rd_valid, self._ren)
            else:
                self.wire(self._rd_valid, self._ren & ~self._wen)

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_dat_to_write(self):
        if ~self._rst_n:
            self._data_to_write = 0
        else:
            self._data_to_write = self._data_in

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def set_addr_to_write(self):
        if ~self._rst_n:
            self._addr_to_write = 0
        else:
            self._addr_to_write = self._wr_addr

    @always_comb
    def decode_wen(self, idx):
        self._wen_to_strg[idx] = \
            self._wen & (self._wr_addr[self.b_a_off + self.bank_width - 1, self.b_a_off] == idx)

    @always_comb
    def decode_ren(self, idx):
        self._ren_to_strg[idx] = \
            self._ren & (self._rd_addr[self.b_a_off + self.bank_width - 1, self.b_a_off] == idx)

    @always_comb
    def set_ren_rmw(self, idx):
        self._ren_to_strg[idx] = \
            ((self._wen | self._ren) & self._read_gate &
             (self._rd_addr[self.b_a_off + self.bank_width - 1, self.b_a_off] == idx))

    @always_comb
    def set_wen_rmw(self, idx):
        self._wen_to_strg[idx] = \
            self._write_gate & (self._addr_to_write[self.b_a_off + self.bank_width - 1, self.b_a_off] == idx)

    @always_comb
    def set_addr_rmw(self, idx):
        self._addr_out[idx] = self._rd_addr[self.mem_addr_width + self.word_width - 1, self.word_width]
        # If we are performing the write
        if self._wen & ~self._write_gate:
            self._addr_out[idx] = self._wr_addr[self.mem_addr_width + self.word_width - 1, self.word_width]
        elif self._write_gate:
            self._addr_out[idx] = \
                self._addr_to_write[self.mem_addr_width + self.word_width - 1, self.word_width]

    @always_comb
    def set_data_combined(self, idx):
        # If the word matches the index, use the data to write
        # (replace the word)
        if self._addr_to_write[self.word_width - 1, 0] == idx:
            self._data_combined[idx] = self._data_to_write
        # Otherwise keep the data
        else:
            self._data_combined[idx] = self._data_from_strg[self._rd_bank][idx]


if __name__ == "__main__":
    stg_dut = StrgRAM()
    verilog(stg_dut, filename="strg_ram.sv",
            additional_passes={"lift config regs": lift_config_reg})



from kratos import *
from functools import reduce
import operator
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.attributes.formal_attr import FormalAttr, FormalSignalConstraint


class AddrGen(Generator):
    '''
    Generate addresses for a single port
    '''
    def __init__(self,
                 iterator_support=6,
                 config_width=16):

        super().__init__(f"addr_gen_{iterator_support}_{config_width}", debug=True)

        # Store local...
        self.iterator_support = iterator_support
        self.config_width = config_width

        # PORT DEFS: begin

        # INPUTS
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        self._restart = self.input("restart", 1)

        self._strides = self.input("strides", self.config_width,
                                   size=self.iterator_support,
                                   packed=True, explicit_array=True)
        self._strides.add_attribute(ConfigRegAttr("Strides of address generator"))
        self._strides.add_attribute(FormalAttr(f"{self._strides.name}", FormalSignalConstraint.SOLVE))

        self._starting_addr = self.input("starting_addr", self.config_width)
        self._starting_addr.add_attribute(ConfigRegAttr("Starting address of address generator"))
        self._starting_addr.add_attribute(FormalAttr(f"{self._starting_addr.name}", FormalSignalConstraint.SOLVE))

        self._step = self.input("step", 1)

        self._mux_sel = self.input("mux_sel", max(clog2(self.iterator_support), 1))

        # OUTPUTS
        # TODO why is this config width instead of address width?
        self._addr_out = self.output("addr_out", self.config_width)

        # PORT DEFS: end

        # LOCAL VARIABLES: begin
        self._strt_addr = self.var("strt_addr", self.config_width)

        self._calc_addr = self.var("calc_addr", self.config_width)

        self._max_value = self.var("max_value", self.iterator_support)

        # LOCAL VARIABLES: end
        # GENERATION LOGIC: begin
        self.wire(self._strt_addr, self._starting_addr)
        self.wire(self._addr_out, self._calc_addr)

        self._current_addr = self.var("current_addr", self.config_width)
        # Calculate address by taking previous calculation and adding the muxed stride
        self.wire(self._calc_addr, self._strt_addr + self._current_addr)

        self.add_code(self.calculate_address)
        # GENERATION LOGIC: end

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def calculate_address(self):
        if ~self._rst_n:
            self._current_addr = 0
        elif self._step:
            # mux_sel as 0 but update means that the machine is resetting.
            if self._restart:
                self._current_addr = 0
            else:
                self._current_addr = self._current_addr + self._strides[self._mux_sel]


if __name__ == "__main__":
    db_dut = AddrGen(iterator_support=6)
    verilog(db_dut, filename="addr_gen.sv")



from lake.models.model import Model


class AggAlignerModel(Model):
    '''
    Model for agg aligner
    '''
    def __init__(self,
                 data_width,
                 max_line_length):

        self.max_line_length = max_line_length
        self.data_width = data_width

        self.config = {}
        self.config["line_length"] = 0

        self.count = 0

    def set_config(self, new_config):
        for key, config_val in new_config.items():
            if key not in self.config:
                AssertionError("Gave bad config...")
            else:
                self.config[key] = config_val

        assert self.config["line_length"] <= self.max_line_length, "Exceeded max line length"

    def interact(self, data_in, valid_in):
        '''
        Returns (data_out, valid, align)
        '''
        if valid_in == 0:
            return (data_in, 0, 0)
        else:
            align = 0
            self.count += 1
            if(self.count == self.config["line_length"]):
                self.count = 0
                align = 1
            return (data_in, 1, align)



from lake.models.model import Model
import kratos as kts


# chain model
class ChainModel(Model):

    def __init__(self,
                 data_width,
                 interconnect_output_ports,
                 chain_idx_bits,
                 enable_chain_output,
                 chain_idx_output):

        # generation parameters
        self.data_width = data_width
        self.interconnect_output_ports = interconnect_output_ports
        self.chain_idx_bits = chain_idx_bits

        # configuration registers passed down from top level
        self.enable_chain_output = enable_chain_output
        self.chain_idx_output = chain_idx_output

    def set_config(self, new_config):
        for key, config_val in new_config.items():
            if key not in self.config:
                AssertionError("Gave bad config...")
            else:
                self.config[key] = config_val

    def interact(self, curr_tile_valid_out, curr_tile_data_out, chain_valid_in, chain_data_in):

        chain_data_out_inter = []
        chain_valid_out_inter = []
        for i in range(self.interconnect_output_ports):
            if chain_valid_in[i] == 0:
                chain_data_out_inter.append(curr_tile_data_out[i])
                chain_valid_out_inter.append(curr_tile_valid_out[i])
            else:
                chain_data_out_inter.append(chain_data_in[i])
                chain_valid_out_inter.append(chain_valid_in[i])

        # all combinational outputs
        # set data_out_tile
        if self.enable_chain_output:
            data_out_tile = chain_data_out_inter
        else:
            data_out_tile = curr_tile_data_out

        # set valid_out_tile
        valid_out_tile = []
        if self.enable_chain_output:
            if not (self.chain_idx_output == 0):
                for i in range(self.interconnect_output_ports):
                    valid_out_tile.append(0)
            else:
                valid_out_tile = chain_valid_out_inter
        else:
            valid_out_tile = curr_tile_valid_out

        # set chain_data_out
        chain_data_out = chain_data_out_inter

        # set chain_valid_out
        chain_valid_out = []
        if (self.chain_idx_output == 0) or \
                (not self.enable_chain_output):
            for i in range(self.interconnect_output_ports):
                chain_valid_out.append(0)
        else:
            chain_valid_out = chain_valid_out_inter

        return chain_data_out, chain_valid_out, data_out_tile, valid_out_tile



from kratos import *
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.attributes.formal_attr import FormalAttr, FormalSignalConstraint

import kratos as kts


class ChainAccessor(Generator):
    def __init__(self,
                 data_width,
                 interconnect_output_ports):
        super().__init__("Chain", debug=True)

        # generator parameters
        self.data_width = data_width
        self.interconnect_output_ports = interconnect_output_ports

        # chain enable configuration register
        self._chain_en = self.input("chain_en", 1)
        self._chain_en.add_attribute(ConfigRegAttr("Signal indicating whether to enable chaining"))
        self._chain_en.add_attribute(FormalAttr(self._chain_en.name, FormalSignalConstraint.SET0))

        # inputs
        self._curr_tile_data_out = self.input("curr_tile_data_out",
                                              self.data_width,
                                              size=self.interconnect_output_ports,
                                              packed=True,
                                              explicit_array=True)

        self._chain_data_in = self.input("chain_data_in",
                                         self.data_width,
                                         size=self.interconnect_output_ports,
                                         packed=True,
                                         explicit_array=True)

        self._accessor_output = self.input("accessor_output",
                                           self.interconnect_output_ports)

        self._data_out_tile = self.output("data_out_tile",
                                          self.data_width,
                                          size=self.interconnect_output_ports,
                                          packed=True,
                                          explicit_array=True)

        self.add_code(self.set_data_out)

    @always_comb
    def set_data_out(self):
        for i in range(self.interconnect_output_ports):
            if self._accessor_output[i]:
                self._data_out_tile[i] = self._curr_tile_data_out[i]
            else:
                if self._chain_en:
                    self._data_out_tile[i] = self._chain_data_in[i]
                else:
                    self._data_out_tile[i] = 0


if __name__ == "__main__":
    dut = ChainAccessor(data_width=16,
                        interconnect_output_ports=2)
    verilog(dut, filename="chain_acc.sv")



from lake.models.model import Model


class RegFIFOModel(Model):
    '''
    Model for agg aligner
    '''
    def __init__(self,
                 data_width,
                 width_mult,
                 depth):

        self.data_width = data_width
        self.width_mult = width_mult
        self.depth = depth

        self.config = {}

        self.rd_ptr = 0
        self.wr_ptr = 0

        self.num_items = 0
        self.reg_array = []
        self.mvd_array = []
        for i in range(self.depth):
            row = []
            for j in range(self.width_mult):
                row.append(0)
            self.reg_array.append(row)
            self.mvd_array.append(row)

        self.full = 0
        self.empty = 1

    # Doesn't have configuration state
    def set_config(self, new_config):
        return

    def get_full(self, push, pop):
        self.full = self.num_items == self.depth
        return self.full

    def get_empty(self, push, pop):
        self.empty = self.num_items == 0
        return self.empty

    def increment_rd(self):
        self.rd_ptr += 1
        if(self.rd_ptr == self.depth):
            self.rd_ptr = 0

    def increment_wr(self):
        self.wr_ptr += 1
        if(self.wr_ptr == self.depth):
            self.wr_ptr = 0

    # Assume full + empty already obeyed
    def interact(self, push, pop, data_in, mem_valid_data):
        '''
        Returns (data_out, valid, empty, full, mem_valid_data_out)
        '''
        empty_ret = int(self.num_items == 0)
        full_ret = int(self.num_items == self.depth)

        if push and pop:
            # Push and pop on empty passes through
            if(self.num_items == 0):
                return (data_in, 1, empty_ret, full_ret, mem_valid_data)
            else:
                dat_out = self.reg_array[self.rd_ptr]
                mem_valid_data_out = self.mvd_array[self.rd_ptr]
                self.increment_rd()
                self.reg_array[self.wr_ptr] = list(data_in)
                self.mvd_array[self.wr_ptr] = mem_valid_data
                self.increment_wr()
                return (dat_out, 1, empty_ret, full_ret, mem_valid_data_out)
        elif push and not pop:
            # Not full, push an item
            if(self.num_items == self.depth):
                return ([0], 0, empty_ret, full_ret, 0)
            self.reg_array[self.wr_ptr] = list(data_in)
            self.mvd_array[self.wr_ptr] = mem_valid_data
            self.increment_wr()
            self.num_items += 1
            # return (self.reg_array[self.rd_ptr], 0, empty_ret, full_ret)
            return ([0], 0, empty_ret, full_ret, 0)
        elif not push and pop:
            if(self.num_items == 0):
                # return (self.reg_array[self.rd_ptr], 0, empty_ret, full_ret)
                return ([0], 0, empty_ret, full_ret, 0)
            dat_out = self.reg_array[self.rd_ptr]
            mem_valid_data_out = self.mvd_array[self.rd_ptr]
            self.increment_rd()
            self.num_items -= 1
            return (dat_out, 1, empty_ret, full_ret, mem_valid_data_out)
        else:
            # return (self.reg_array[self.rd_ptr], 0, empty_ret, full_ret)
            return ([0], 0, empty_ret, full_ret, 0)



from kratos import *
from math import log


class TwoPortSRAMStub(Generator):
    '''
    2 port SRAM
    '''
    ##########################
    # Generation             #
    ##########################
    def __init__(self, width, depth, bypass):
        super().__init__("two_port_sram_stub", True)

        ############################
        # Clock and Reset          #
        ############################
        self.i_clk = self.clock("i_clk")
        self.i_rst_n = self.reset("i_rst_n", 1)

        ############################
        # Inputs                   #
        ############################
        self.i_wen = self.input("i_wen", 1)
        self.i_ren = self.input("i_ren", 1)

        self.i_wr_addr = self.input("i_wr_addr", clog2(depth))
        self.i_rd_addr = self.input("i_rd_addr", clog2(depth))

        self.i_data = self.input("i_data", width)

        ############################
        # Outputs                  #
        ############################
        self.o_data = self.output("o_data", width)

        ############################
        # Local Variables          #
        ############################
        self.data_array = self.var("data_array", width=width, size=depth, packed=True)

        self.bypass = bypass

        ############################
        # Add seq blocks           #
        ############################
        self.add_code(self.seq_data_access)
        self.add_code(self.seq_data_out)

    ##########################
    # Access sram array      #
    ##########################
    @always((posedge, "i_clk"))
    def seq_data_access(self):
        if(self.i_wen):
            self.data_array[self.i_wr_addr] = self.i_data

    @always((posedge, "i_clk"), (negedge, "i_rst_n"))
    def seq_data_out(self):
        if ~self.i_rst_n:
            self.o_data = 0
        elif self.i_ren:
            if (self.i_rd_addr == self.i_wr_addr) & (self.bypass == 1):
                self.o_data = self.i_data
            else:
                self.o_data = self.data_array[self.i_rd_addr]


if __name__ == "__main__":
    dut = TwoPortSRAMStub(16, 1024, 0)
    verilog(dut, filename="two_port_sram_stub.sv", check_active_high=False)



from kratos import *
from lake.modules.passthru import *
from lake.modules.register_file import RegisterFile
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.attributes.range_group import RangeGroupAttr
from lake.passes.passes import lift_config_reg
from lake.modules.sram_stub import SRAMStub
from lake.modules.for_loop import ForLoop
from lake.modules.addr_gen import AddrGen
from lake.modules.spec.sched_gen import SchedGen
import kratos as kts


class StrgUBThin(Generator):
    def __init__(self,
                 data_width=16,  # CGRA Params
                 mem_width=16,
                 mem_depth=512,
                 banks=1,
                 input_addr_iterator_support=6,
                 input_sched_iterator_support=6,
                 output_addr_iterator_support=6,
                 output_sched_iterator_support=6,
                 interconnect_input_ports=1,  # Connection to int
                 interconnect_output_ports=1,
                 config_width=16,
                 mem_input_ports=1,
                 mem_output_ports=1,
                 read_delay=1,  # Cycle delay in read (SRAM vs Register File)
                 rw_same_cycle=True,
                 gen_addr=True):

        super().__init__("strg_ub_thin", debug=True)

        assert mem_width == data_width, f"This module should only be used when the fetch width is 1!"

        self.fetch_width = mem_width // data_width
        self.interconnect_input_ports = interconnect_input_ports
        self.interconnect_output_ports = interconnect_output_ports
        self.mem_depth = mem_depth
        self.config_width = config_width
        self.data_width = data_width
        self.input_addr_iterator_support = input_addr_iterator_support
        self.input_sched_iterator_support = input_sched_iterator_support
        self.output_addr_iterator_support = output_addr_iterator_support
        self.output_sched_iterator_support = output_sched_iterator_support
        self.rw_same_cycle = rw_same_cycle
        self.read_delay = read_delay
        self.gen_addr = gen_addr
        self.default_iterator_support = 6
        self.default_config_width = 16
        # generation parameters
        # inputs
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        self._data_in = self.input("data_in", self.data_width,
                                   size=self.interconnect_input_ports,
                                   packed=True,
                                   explicit_array=True)

        # outputs
        self._data_out = self.output("data_out", self.data_width,
                                     size=self.interconnect_output_ports,
                                     packed=True,
                                     explicit_array=True)

        self._data_to_sram = self.output("data_to_strg", self.data_width,
                                         size=self.fetch_width,
                                         packed=True)
        self._data_from_sram = self.input("data_from_strg", self.data_width,
                                          size=self.fetch_width,
                                          packed=True)
        # Early out in case...
        if self.gen_addr is False:
            # Pass through write enable, addr data and
            # read enable, addr data
            self._read = self.input("ren_in", 1)
            self._write = self.input("wen_in", 1)
            self._write_addr = self.input("write_addr", self.config_width)
            self._read_addr = self.input("read_addr", self.config_width)
            self._cen_to_sram = self.output("cen_to_strg", 1, packed=True)
            self._wen_to_sram = self.output("wen_to_strg", 1, packed=True)
            self._ren_to_sram = self.output("ren_to_strg", 1, packed=True)
            self._wr_addr_to_sram = self.output("wr_addr_out", clog2(self.mem_depth), packed=True)
            self._rd_addr_to_sram = self.output("rd_addr_out", clog2(self.mem_depth), packed=True)
            self._accessor_output = self.output("accessor_output", self.interconnect_output_ports)
            self.wire(self._accessor_output, self._read)
            self.wire(self._cen_to_sram, self._write | self._read)
            self.wire(self._wen_to_sram, self._write)
            self.wire(self._ren_to_sram, self._read)
            self.wire(self._data_out, self._data_from_sram)
            self.wire(self._data_in, self._data_to_sram)
            self.wire(self._wr_addr_to_sram, self._write_addr[clog2(self.mem_depth) - 1, 0])
            self.wire(self._rd_addr_to_sram, self._read_addr[clog2(self.mem_depth) - 1, 0])

            return

        # Create cycle counter to share...
        self._cycle_count = self.var("cycle_count", 16)
        self.add_code(self.increment_cycle_count)

        # local variables
        self._write = self.var("write", 1)
        self._read = self.var("read", self.interconnect_output_ports)
        self._accessor_output = self.output("accessor_output", self.interconnect_output_ports)
        self.wire(self._accessor_output, self._read)

        self._valid_out = self.output("valid_out", 1)
        if self.read_delay == 1:
            self._read_d1 = self.var("read_d1", 1)
            self.add_code(self.delay_read)
            self.wire(self._valid_out, self._read_d1)
        else:
            self.wire(self._valid_out, self._read)

        self._write_addr = self.var("write_addr", self.config_width)
        self._read_addr = self.var("read_addr", self.config_width)
        self._addr = self.var("addr", clog2(self.mem_depth))

        # Create for loop counters that can be shared across the input port selection and SRAM write
        fl_ctr_sram_wr = ForLoop(iterator_support=self.default_iterator_support,
                                 config_width=self.default_config_width)
        loop_itr = fl_ctr_sram_wr.get_iter()
        loop_wth = fl_ctr_sram_wr.get_cfg_width()

        self.add_child(f"sram_write_loops",
                       fl_ctr_sram_wr,
                       clk=self._clk,
                       rst_n=self._rst_n,
                       step=self._write)

        # Whatever comes through here should hopefully just pipe through seamlessly
        # addressor modules
        self.add_child(f"sram_write_addr_gen",
                       AddrGen(iterator_support=self.default_iterator_support,
                               config_width=self.default_config_width),
                       clk=self._clk,
                       rst_n=self._rst_n,
                       step=self._write,
                       mux_sel=fl_ctr_sram_wr.ports.mux_sel_out,
                       restart=fl_ctr_sram_wr.ports.restart,
                       addr_out=self._write_addr)

        # scheduler modules
        self.add_child(f"sram_write_sched_gen",
                       SchedGen(iterator_support=self.default_iterator_support,
                                config_width=self.default_config_width),
                       clk=self._clk,
                       rst_n=self._rst_n,
                       cycle_count=self._cycle_count,
                       mux_sel=fl_ctr_sram_wr.ports.mux_sel_out,
                       finished=fl_ctr_sram_wr.ports.restart,
                       valid_output=self._write)

        # -------------------------------- Delineate new group -------------------------------
        fl_ctr_sram_rd = ForLoop(iterator_support=self.default_iterator_support,
                                 config_width=self.default_config_width)
        loop_itr = fl_ctr_sram_rd.get_iter()
        loop_wth = fl_ctr_sram_rd.get_cfg_width()

        self.add_child(f"sram_read_loops",
                       fl_ctr_sram_rd,
                       clk=self._clk,
                       rst_n=self._rst_n,
                       step=self._read)

        self.add_child(f"sram_read_addr_gen",
                       AddrGen(iterator_support=self.default_iterator_support,
                               config_width=self.default_config_width),
                       clk=self._clk,
                       rst_n=self._rst_n,
                       step=self._read,
                       mux_sel=fl_ctr_sram_rd.ports.mux_sel_out,
                       restart=fl_ctr_sram_rd.ports.restart,
                       addr_out=self._read_addr)

        self.add_child(f"sram_read_sched_gen",
                       SchedGen(iterator_support=self.default_iterator_support,
                                config_width=self.default_config_width),
                       clk=self._clk,
                       rst_n=self._rst_n,
                       cycle_count=self._cycle_count,
                       mux_sel=fl_ctr_sram_rd.ports.mux_sel_out,
                       finished=fl_ctr_sram_rd.ports.restart,
                       valid_output=self._read)

        # Now deal with dual_port/single_port madness...
        self._cen_to_sram = self.output("cen_to_strg", 1, packed=True)
        self._wen_to_sram = self.output("wen_to_strg", 1, packed=True)
        self._ren_to_sram = self.output("ren_to_strg", 1, packed=True)
        self.wire(self._cen_to_sram, self._write | self._read)
        self.wire(self._wen_to_sram, self._write)
        self.wire(self._ren_to_sram, self._read)
        self.wire(self._data_out, self._data_from_sram)
        self.wire(self._data_in, self._data_to_sram)

        if self.rw_same_cycle:
            # If we can read and write the same cycle we
            # can pretty safeuly assume we have separate read/write ports...
            self._wr_addr_to_sram = self.output("wr_addr_out", clog2(self.mem_depth), packed=True)
            self._rd_addr_to_sram = self.output("rd_addr_out", clog2(self.mem_depth), packed=True)
            self.wire(self._wr_addr_to_sram, self._write_addr[clog2(self.mem_depth) - 1, 0])
            self.wire(self._rd_addr_to_sram, self._read_addr[clog2(self.mem_depth) - 1, 0])
        else:
            self._addr_to_sram = self.output("addr_out", clog2(self.mem_depth), packed=True)
            self.wire(self._addr_to_sram, self._addr)
            self.add_code(self.set_sram_addr)

    @always_comb
    def set_sram_addr(self):
        if self._write:
            self._addr = self._write_addr[clog2(self.mem_depth) - 1, 0]
        else:
            self._addr = self._read_addr[clog2(self.mem_depth) - 1, 0]

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def delay_read(self):
        if ~self._rst_n:
            self._read_d1 = 0
        else:
            self._read_d1 = self._read

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def increment_cycle_count(self):
        if ~self._rst_n:
            self._cycle_count = 0
        else:
            self._cycle_count = self._cycle_count + 1


if __name__ == "__main__":
    lake_dut = StrgUBThin()
    verilog(lake_dut, filename="strg_ub_thin.sv",
            optimize_if=False,
            additional_passes={"lift config regs": lift_config_reg})



from kratos import *
from math import log
from lake.modules.sram_stub import SRAMStub
from lake.modules.sram_stub_generator import SRAMStubGenerator
from lake.utils.flattenND import FlattenND
from lake.utils.reverse_flatten import ReverseFlatten


class SRAM(Generator):

    def __init__(self,
                 use_sram_stub,
                 sram_name,
                 data_width,
                 fw_int,
                 mem_depth,
                 mem_input_ports,
                 mem_output_ports,
                 address_width,
                 bank_num,
                 num_tiles):
        super().__init__(f"{sram_name}_generator")

        self.use_sram_stub = use_sram_stub
        self.sram_name = sram_name
        self.data_width = data_width
        self.fw_int = fw_int
        self.mem_depth = mem_depth
        self.mem_input_ports = mem_input_ports
        self.mem_output_ports = mem_output_ports
        self.address_width = address_width
        self.bank_num = bank_num
        self.num_tiles = num_tiles

        self.chain_idx_bits = max(1, clog2(num_tiles))

        self._gclk = self.clock("clk")

        self._clk_en = self.input("clk_en", 1)

        if self.fw_int > 1:
            self._mem_data_in_bank = self.input("mem_data_in_bank",
                                                self.data_width,
                                                size=self.fw_int,
                                                packed=True,
                                                explicit_array=True)

            self._mem_data_out_bank = self.output("mem_data_out_bank",
                                                  self.data_width,
                                                  size=(self.mem_output_ports,
                                                        self.fw_int),
                                                  packed=True,
                                                  explicit_array=True)

            if not self.use_sram_stub:
                # flattened version of input data
                self._sram_mem_data_in_bank = self.var("sram_mem_data_in_bank",
                                                       self.data_width *
                                                       self.fw_int *
                                                       self.mem_input_ports)

                # flattened version of output data
                self._sram_mem_data_out_bank = self.var("sram_mem_data_out_bank",
                                                        self.data_width *
                                                        self.fw_int *
                                                        self.mem_output_ports)
        else:
            self._mem_data_in_bank = self.input("mem_data_in_bank", self.data_width, packed=True)

            self._mem_data_out_bank = self.output("mem_data_out_bank", self.data_width, packed=True)

        self._mem_addr_in_bank = self.input("mem_addr_in_bank",
                                            self.address_width)
        if num_tiles == 1:
            self._mem_addr_to_sram = self.var("mem_addr_to_sram",
                                              self.address_width)
        else:
            self._mem_addr_to_sram = self.var("mem_addr_to_sram",
                                              self.address_width - self.chain_idx_bits)

        self._mem_cen_in_bank = self.input("mem_cen_in_bank", self.mem_output_ports)

        self._mem_wen_in_bank = self.input("mem_wen_in_bank", self.mem_input_ports)

        self._wtsel = self.input("wtsel", 2)
        self._rtsel = self.input("rtsel", 2)

        if self.use_sram_stub:
            mbank = SRAMStub(data_width=self.data_width,
                             width_mult=self.fw_int,
                             depth=self.mem_depth)

            self.add_child(f"mem_{self.bank_num}", mbank,
                           clk=self._gclk,
                           data_in=self._mem_data_in_bank,
                           addr=self._mem_addr_to_sram,
                           cen=self._mem_cen_in_bank,
                           wen=self._mem_wen_in_bank,
                           data_out=self._mem_data_out_bank)

        # instantiante external provided sram macro and flatten input/output data
        # if fetch width is greater than 1
        else:
            mbank = SRAMStubGenerator(sram_name=self.sram_name,
                                      data_width=self.data_width,
                                      width_mult=self.fw_int,
                                      depth=self.mem_depth)

            compose_wide = True

            if self.fw_int > 1:
                flatten_data_in = FlattenND(self.data_width,
                                            self.fw_int,
                                            self.mem_input_ports)

                self.add_child(f"flatten_data_in_{self.bank_num}",
                               flatten_data_in,
                               input_array=self._mem_data_in_bank,
                               output_array=self._sram_mem_data_in_bank)

                if compose_wide:
                    for j in range(2):
                        mbank = SRAMStubGenerator(sram_name=self.sram_name,
                                                  data_width=self.data_width,
                                                  width_mult=self.fw_int // 2,
                                                  depth=self.mem_depth)

                        self.add_child(f"mem_inst_{self.bank_num}_pt_{j}",
                                       mbank,
                                       sram_addr=self._mem_addr_to_sram,
                                       sram_cen=~self._mem_cen_in_bank,
                                       sram_clk=self._gclk,
                                       sram_data_in=self._sram_mem_data_in_bank[j * 32 + 32 - 1, j * 32],
                                       sram_data_out=self._sram_mem_data_out_bank[j * 32 + 32 - 1, j * 32],
                                       sram_wen=~self._mem_wen_in_bank,
                                       sram_wtsel=self._wtsel,
                                       sram_rtsel=self._rtsel)

                else:
                    self.add_child(f"mem_inst_{self.bank_num}",
                                   mbank,
                                   sram_addr=self._mem_addr_to_sram,
                                   sram_cen=~self._mem_cen_in_bank,
                                   sram_clk=self._gclk,
                                   sram_data_in=self._sram_mem_data_in_bank,
                                   sram_data_out=self._sram_mem_data_out_bank,
                                   sram_wen=~self._mem_wen_in_bank,
                                   sram_wtsel=self._wtsel,
                                   sram_rtsel=self._rtsel)

                flatten_data_out = ReverseFlatten(self.data_width,
                                                  self.fw_int,
                                                  self.mem_output_ports)

                self.add_child(f"flatten_data_out_{self.bank_num}",
                               flatten_data_out,
                               input_array=self._sram_mem_data_out_bank,
                               output_array=self._mem_data_out_bank)

            else:
                self.add_child(f"mem_inst_{self.bank_num}",
                               mbank,
                               sram_addr=self._mem_addr_to_sram,
                               sram_cen=~self._mem_cen_in_bank,
                               sram_clk=self._gclk,
                               sram_data_in=self._mem_data_in_bank,
                               sram_data_out=self._mem_data_out_bank,
                               sram_wen=~self._mem_wen_in_bank,
                               sram_wtsel=self._wtsel,
                               sram_rtsel=self._rtsel)

        self.add_code(self.set_mem_addr)

    @always_comb
    def set_mem_addr(self):
        # these ranges are inclusive
        if self.num_tiles == 1:
            self._mem_addr_to_sram = self._mem_addr_in_bank
        else:
            self._mem_addr_to_sram = self._mem_addr_in_bank[self.address_width - self.chain_idx_bits - 1, 0]


if __name__ == "__main__":
    dut = SRAM(use_sram_stub=False,
               sram_name="TSMC",
               data_width=16,
               fw_int=4,
               mem_depth=128,
               mem_input_ports=1,
               mem_output_ports=1,
               address_width=7,
               bank_num=4,
               num_tiles=1)
    verilog(dut, filename="wrapper.sv")



from lake.models.model import Model


class DemuxReadsModel(Model):
    '''
    Model for agg aligner
    '''
    def __init__(self,
                 fetch_width,
                 data_width,
                 banks,
                 int_out_ports):

        self.fetch_width = fetch_width
        self.data_width = data_width
        self.fw_int = int(fetch_width / data_width)
        self.banks = banks
        self.int_out_ports = int_out_ports

    # Doesn't actually have configuration state
    def set_config(self, new_config):
        return

    def interact(self, data_in, valid_in, port_in, mem_valid_data):
        '''
        Returns (data_out, valid_out, mem_valid_data_out)
        '''
        data_out = []
        valid_out = []
        mem_valid_data_out = []
        for i in range(self.int_out_ports):
            row = []
            for j in range(self.fw_int):
                row.append(0)
            data_out.append(list(row))
            valid_out.append(0)
            mem_valid_data_out.append(0)

        no_valid = True
        for i in range(self.banks):
            if(valid_in[i]):
                no_valid = False

        if no_valid:
            return (data_out, valid_out, mem_valid_data_out)

        for i in range(self.int_out_ports):
            for j in range(self.banks):
                if(valid_in[j] & (port_in[j] == (1 << i))):
                    data_out[i] = list(data_in[j])
                    valid_out[i] = 1
                    mem_valid_data_out[i] = mem_valid_data[j]
                    break

        return (data_out, valid_out, mem_valid_data_out)



from kratos import *
import kratos as kts
from _kratos import create_wrapper_flatten
from math import log

from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.attributes.formal_attr import FormalAttr, FormalSignalConstraint
from lake.dsl.edge import get_full_edge_params
from lake.dsl.helper import *
from lake.dsl.mem_port import MemPort
from lake.dsl.memory import mem_inst
from lake.modules.for_loop import ForLoop
from lake.modules.addr_gen import AddrGen
from lake.modules.spec.sched_gen import SchedGen
from lake.passes.passes import lift_config_reg
from lake.utils.util import safe_wire, trim_config_list
from lake.utils.util import extract_formal_annotation, modular_formal_annotation
from lake.utils.parse_clkwork_config import *


'''
    Basically creating configuration wrapper with
    bogus config bus for synthesis numbers.
'''


class CFGRegWrapper(Generator):
    def __init__(self,
                 child_gen: kts.Generator):

        super().__init__(f"{child_gen.name}_cfg_W", debug=True)

        # Works for now since we always have these...
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        self.add_child_generator(f"{child_gen.name}", child_gen,
                                 clk=self._clk,
                                 rst_n=self._rst_n)

        # Create fake config bus...
        self._config_en = self.input("config_en", 1)
        self._config_data = self.input("config_data", 256)

        # Now copy inputs and outputs that aren't config regs

        for port in child_gen.ports:
            actual_port = child_gen.ports[port]
            self.uplevel_port(actual_port)

    def uplevel_port(self, port):

        pname = port.name
        pdir = str(port.port_direction)
        ptype = str(port.port_type)
        pattrs = port.get_attributes()

        if pname == "clk" or pname == "rst_n":
            return

        if ptype == "PortType.Data" and pdir == "PortDirection.In":
            cfgreg_attr = False
            for attr in pattrs:
                if isinstance(attr, ConfigRegAttr):
                    cfgreg_attr = True
            # We know to hook it up to bogus config bus
            if cfgreg_attr is True:
                newvar = self.var_from_def(port, pname)
                self.wire(newvar, port)

                @always_ff((posedge, "clk"), (negedge, "rst_n"))
                def add_to_config_reg(self):
                    if ~self._rst_n:
                        newvar = 0
                    elif self._config_en:
                        newvar = 1

                self.add_code(add_to_config_reg)
            # Otherwise pass it through...
            else:
                newin = self.port_from_def(port, pname)
                self.wire(newin, port)
        else:
            newport = self.port_from_def(port, pname)
            self.wire(newport, port)



from kratos import *


class LineBufferControl(Generator):
    def __init__(self):
        super().__init__("linebuffer_control", True)

        # PORT DEFS: begin
        self._clk = self.clock("clk", 1)
        self._clk_en = self.input("clk_en", 1)
        self._reset = self.reset("reset")
        self._flush = self.input("flush", 1)

        self._wen = self.input("wen", 1)
        self._depth = self.input("depth", 16)
        self._stencil_width = self.input("stencil_width", 16)
        self._num_words_mem = self.input("num_words_mem", 16)

        self._valid = self.output("valid", 1)
        self._ren_to_fifo = self.output("ren_to_fifo", 1)
        # PORT DEFS: end

        # LOCAL SIGNALS: begin
        self._vg_ctr = self.var("vg_ctr", 16)
        self._valid_gate = self.var("valid_gate", 1)
        self._valid_int = self.var("valid_int", 1)
        self._threshold = self.var("threshold", 1)
        # LOCAL SIGNALS: end

        # GENERATION LOGIC: begin
        self.wire(self._valid_gate,
                  ternary((self._stencil_width == const(0, 16)),
                          const(1, 1),
                          (self._vg_ctr > (self._stencil_width - const(1, 16)))))
        self.wire(self._valid_int,
                  (self._num_words_mem >= (self._depth - 1)) &
                  self._wen & self._threshold & (self._depth > 0))
        self.wire(self._valid, self._valid_gate & self._valid_int)
        self.wire(self._ren_to_fifo, (self._num_words_mem >= (self._depth - 1)) &
                  self._wen & (self._depth > 0))

        self.add_code(self.threshold_update)
        self.add_code(self.vg_ctr_update)
        # GENERATION LOGIC: end

    @always((posedge, "clk"), (posedge, "reset"))
    def threshold_update(self):
        if(self._reset):
            self._threshold = 0
        elif self._clk_en:
            if self._flush:
                self._threshold = 0
            elif (self._num_words_mem == (self._depth - 1)) & self._wen:
                self._threshold = 1

    @always((posedge, "clk"), (posedge, "reset"))
    def vg_ctr_update(self):
        if(self._reset):
            self._vg_ctr = 0
        elif self._clk_en:
            if self._flush:
                self._vg_ctr = 0
            elif self._valid_int:
                if self._vg_ctr == (self._depth - 1):
                    self._vg_ctr = 0
                else:
                    self._vg_ctr = self._vg_ctr + 1


if __name__ == "__main__":
    dut = LineBufferControl()
    verilog(dut, filename="linebuffer_control.sv", check_active_high=False)



from kratos import *
from functools import reduce
import operator


class DoubleBufferControl(Generator):
    def __init__(self,
                 data_width,
                 mem_depth,
                 banks,
                 iterator_support):
        super().__init__("doublebuffer_control")

        self.data_width = data_width
        self.banks = banks
        self.mem_depth = mem_depth
        self.mem_addr_width = clog2(self.mem_depth)
        self.bank_addr_width = clog2(self.banks)
        self.full_addr = self.mem_addr_width + self.bank_addr_width
        self.iterator_support = iterator_support

        # PORT DEFS: begin
        self._clk = self.clock("clk")
        self._clk_en = self.input("clk_en", 1)
        self._reset = self.reset("reset")
        self._flush = self.input("flush", 1)

        self._ren = self.input("ren", 1)
        self._wen = self.input("wen", 1)
        self._data_in = self.input("data_in", self.data_width)
        self._data_out = self.output("data_out", self.data_width)
        self._addr_in = self.input("addr_in", self.data_width)

        self._doublebuffer_data_in = self.output("doublebuffer_data_in",
                                                 self.data_width, size=self.banks,
                                                 explicit_array=True,
                                                 packed=True)
        self._doublebuffer_cen_mem = self.output("doublebuffer_cen_mem", self.banks)
        self._doublebuffer_wen_mem = self.output("doublebuffer_wen_mem", self.banks)
        self._doublebuffer_addr_mem = self.output("doublebuffer_addr_mem",
                                                  self.mem_addr_width,
                                                  size=self.banks,
                                                  explicit_array=True,
                                                  packed=True)
        self._doublebuffer_data_out = self.input("doublebuffer_data_out",
                                                 self.data_width,
                                                 size=self.banks,
                                                 explicit_array=True,
                                                 packed=True)

        self._depth = self.input("depth", 16)
        self._valid = self.output("valid", 1)
        self._switch = self.input("switch", 1)

        self._chain_idx = self.input("chain_idx", 4)

        self._arbitrary_addr = self.input("arbitrary_addr", 1)
        self._starting_addr = self.input("starting_addr", 16)
        self._iter_cnt = self.input("iter_cnt", 32)
        self._dimensionality = self.input("dimensionality", 4)

        self._stride = self.input("stride", 16,
                                  size=self.iterator_support,
                                  packed=True,
                                  explicit_array=True)
        self._range = self.input("range", 32,
                                 size=self.iterator_support,
                                 packed=True,
                                 explicit_array=True)

        self._rate_matched = self.input("rate_matched", 1)
        self._stencil_width = self.input("stencil_width", 32)
        # PORT DEFS: end

        # LOCAL VARIABLES: begin
        self._current_loc = self.var("current_loc", 16,
                                     size=self.iterator_support,
                                     packed=True,
                                     explicit_array=True)
        self._valid_arb = self.var("valid_arb", 1)
        self._init_state = self.var("init_state", 1)
        self._addr = self.var("addr", self.mem_addr_width)
        self._ping_npong = self.var("ping_npong", 1)
        self._read_addr = self.var("read_addr", 16)
        self._write_addr = self.var("write_addr", 16)
        self._dim_counter = self.var("dim_counter", 32,
                                     size=self.iterator_support,
                                     packed=True,
                                     explicit_array=True)

        self._update = self.var("update", self.iterator_support)
        self._strt_addr = self.var("strt_addr", 16)
        self._read_cnt = self.var("read_cnt", 32)
        self._firstn = self.var("firstn", self.data_width,
                                size=self.banks,
                                explicit_array=True,
                                packed=True)
        self._take_the_flop = self.var("take_the_flop", 1)
        self._autoswitch = self.var("autoswitch", 1)
        self._read_done = self.var("read_done", 1)
        self._read_done_thresh = self.var("read_done_thresh", 1)
        self._write_done = self.var("write_done", 1)
        self._write_done_thresh = self.var("write_done_thresh", 1)

        self._last_line_gate = self.var("last_line_gate", 1)
        self._read_first = self.var("read_first", 1)
        self._next_take_the_flop = self.var("next_take_the_flop", 1)
        self._write_in_range = self.var("write_in_range", 1)
        self._read_in_range = self.var("read_in_range", 1)
        self._read_in_range_d1 = self.var("read_in_range_d1", 1)
        self._read_mux = self.var("read_mux", 1)
        self._counter_update = self.var("counter_update", 1)
        self._calc_addr = self.var("calc_addr", 16)
        # LOCAL VARIABLES: end

        # GENERATION LOGIC: begin
        self.wire(self._counter_update,
                  ((self._init_state | (self._depth == 0)) &
                   self._read_mux & ~self._read_done &
                   ~self._take_the_flop & ~self._read_done_thresh) |
                  (self._init_state & self._read_mux & self._take_the_flop))

        self.wire(self._read_mux,
                  ternary(self._rate_matched,
                          self._wen,
                          self._ren))
        self.wire(self._autoswitch,
                  ~self._arbitrary_addr &
                  (self._write_done | self._write_done_thresh) &
                  (self._read_done | self._read_done_thresh | ~self._init_state) &
                  ~(self._depth == const(0, self._depth.width)))

        self.wire(self._strt_addr, self._starting_addr[self.data_width - 1, 0])
        self.wire(self._addr, self._addr_in[self.mem_addr_width - 1, 0])
        self.wire(self._last_line_gate,
                  ternary(self._stencil_width == 0,
                          const(1, 1),
                          (self._read_cnt >= (self._stencil_width - 1))))

        # Figure out when read iterations are done
        self.wire(self._read_done, (self._read_cnt == (self._iter_cnt - 1)) & self._read_mux)
        self.add_code(self.read_done_thresh_update)

        # Figure out when write iterations are done
        self.wire(self._write_done, (self._write_addr == (self._depth - 1)) & self._wen)
        self.add_code(self.write_done_thresh_update)

        # Check that reads/writes are in the range of the modules
        self.wire(self._write_in_range,
                  self._write_addr[self.mem_addr_width + 4 - 1,
                                   self.mem_addr_width] == self._chain_idx)
        self.wire(self._read_in_range,
                  self._read_addr[self.mem_addr_width + 4 - 1,
                                  self.mem_addr_width] == self._chain_idx)

        # Set valid to be this junk of logic
        self.wire(self._valid,
                  ternary(self._arbitrary_addr,
                          self._valid_arb,
                          self._last_line_gate &
                          self._read_mux &
                          (self._init_state | (self._depth == 0)) &
                          self._read_in_range_d1 & ~self._read_done_thresh))

        self.add_code(self.valid_arb_update)
        self.add_code(self.read_in_range_d1_update)

        for i in range(self.banks):
            self.wire(self._doublebuffer_data_in[i], self._data_in)

            self.wire(self._doublebuffer_cen_mem[i],
                      (self._wen & (self._ping_npong == i)) |
                      self._flush |
                      self._switch |
                      self._autoswitch |
                      (self._read_mux & ~(self._ping_npong == i)))

            self.wire(self._doublebuffer_wen_mem[i],
                      (self._ping_npong == i) &
                      (self._wen & ~self._write_done_thresh) &
                      self._write_in_range &
                      ~(self._depth == 0))

            self.wire(self._doublebuffer_addr_mem[i],
                      ternary(self._ping_npong == i,
                              self._write_addr[self.mem_addr_width - 1, 0],
                              self._read_addr[self.mem_addr_width - 1, 0]))

        self.wire(self._data_out,
                  ternary(self._take_the_flop,
                          self._firstn[~self._ping_npong],
                          self._doublebuffer_data_out[~self._ping_npong]))

        # Set read address
        self.wire(self._read_addr,
                  ternary(self._arbitrary_addr,
                          concat(const(0, 7), self._addr),
                          self._calc_addr))

        self.add_code(self.calc_addr_comb)

        # Set update vector
        self.wire(self._update[0], self._init_state | (self._depth == 0))
        for i in range(self.iterator_support - 1):
            self.wire(self._update[i + 1],
                      (self._dim_counter[i] == (self._range[i] - 1)) & self._update[i])

        # Handle the prebuffer on the first element to a bank
        self.add_code(self.firstn_update)

        self.wire(self._next_take_the_flop,
                  ternary(self._autoswitch,
                          const(1, 1),
                          ternary(self._take_the_flop & ~self._read_mux,
                                  const(1, 1),
                                  const(0, 1))))

        self.add_code(self.read_first_update)
        self.add_code(self.take_the_flop_update)
        self.add_code(self.init_state_update)

        self.add_code(self.dim_counter_update)
        self.add_code(self.current_loc_update)
        self.add_code(self.ping_npong_update)

        self.add_code(self.write_addr_update)
        self.add_code(self.read_cnt_update)

        # GENERATION LOGIC: end

    def calc_addr_comb(self):
        self._calc_addr = reduce(operator.add,
                                 [(ternary(const(i, self._dimensionality.width) <
                                           self._dimensionality,
                                           self._current_loc[i],
                                           const(0, self._calc_addr.width)))
                                  for i in range(self.iterator_support)] + [self._strt_addr])

    @always((posedge, "clk"), (posedge, "reset"))
    def read_done_thresh_update(self):
        if self._reset:
            self._read_done_thresh = 0
        elif self._clk_en:
            if self._flush:
                self._read_done_thresh = 0
            else:
                if self._autoswitch | self._switch:
                    self._read_done_thresh = 0
                elif self._read_done:
                    self._read_done_thresh = 1

    @always((posedge, "clk"), (posedge, "reset"))
    def write_done_thresh_update(self):
        if self._reset:
            self._write_done_thresh = 0
        elif self._clk_en:
            if self._flush:
                self._write_done_thresh = 0
            else:
                if self._autoswitch | self._switch:
                    self._write_done_thresh = 0
                elif self._write_done:
                    self._write_done_thresh = 1

    @always((posedge, "clk"), (posedge, "reset"))
    def valid_arb_update(self):
        if self._reset:
            self._valid_arb = 0
        elif self._clk_en:
            if self._flush | self._switch:
                self._valid_arb = 0
            else:
                self._valid_arb = self._read_in_range & self._read_mux & self._init_state

    @always((posedge, "clk"), (posedge, "reset"))
    def read_in_range_d1_update(self):
        if self._reset:
            self._read_in_range_d1 = 0
        elif self._clk_en:
            # Handle some weird zero-depth stuff
            if self._flush:
                self._read_in_range_d1 = self._strt_addr[self.mem_addr_width + 4 - 1,
                                                         self.mem_addr_width] == self._chain_idx
            else:
                self._read_in_range_d1 = self._read_in_range

    def read_addr_comb(self):
        return 0

    def update_comb(self):
        return 0

    @always((posedge, "clk"), (posedge, "reset"))
    def firstn_update(self):
        if self._reset:
            self._firstn = 0
        elif self._clk_en:
            if self._flush:
                self._firstn = 0
            elif self._wen & (self._write_addr == self._strt_addr):
                # TODO: MAKE THE LOGIC GENERAL
                self._firstn[0] = ternary(~self._ping_npong, self._data_in, self._firstn[0])
                self._firstn[1] = ternary(self._ping_npong, self._data_in, self._firstn[1])

    @always((posedge, "clk"), (posedge, "reset"))
    def read_first_update(self):
        if self._reset:
            self._read_first = 1
        elif self._clk_en:
            if self._flush:
                self._read_first = 1
            elif self._take_the_flop & self._read_mux:
                self._read_first = 0

    @always((posedge, "clk"), (posedge, "reset"))
    def take_the_flop_update(self):
        if self._reset:
            self._take_the_flop = 0
        elif self._clk_en:
            if self._flush:
                self._take_the_flop = 0
            else:
                self._take_the_flop = self._next_take_the_flop

    @always((posedge, "clk"), (posedge, "reset"))
    def init_state_update(self):
        if self._reset:
            self._init_state = 0
        elif self._clk_en:
            if self._flush:
                self._init_state = 0
            elif self._autoswitch | self._switch:
                self._init_state = 1

    @always((posedge, "clk"), (posedge, "reset"))
    def dim_counter_update(self):
        if self._reset:
            self._dim_counter = 0
        elif self._clk_en:
            if self._flush:
                self._dim_counter[0] = ternary(self._depth == 0,
                                               concat(const(0, 31), self._range[0] > 1),
                                               const(0, self._dim_counter[0].width))
                for i in range(self.iterator_support - 1):
                    self._dim_counter[i + 1] = 0
            elif self._autoswitch | self._switch:
                self._dim_counter[0] = concat(const(0, 31), self._range[0] > 1)
                for i in range(self.iterator_support - 1):
                    self._dim_counter[i + 1] = 0
            elif self._counter_update & (i < self._dimensionality):
                for i in range(self.iterator_support):
                    if self._update[i]:
                        if self._dim_counter[i] == (self._range[i] - 1):
                            self._dim_counter[i] = 0
                        else:
                            self._dim_counter[i] = self._dim_counter[i] + 1

    @always((posedge, "clk"), (posedge, "reset"))
    def current_loc_update(self):
        if self._reset:
            self._current_loc = 0
        elif self._clk_en:
            if self._flush:
                self._current_loc[0] = ternary(self._depth == 0,
                                               self._stride[0],
                                               const(0, self._current_loc[0].width))
                for i in range(self.iterator_support - 1):
                    self._current_loc[i + 1] = 0
            elif self._autoswitch | self._switch:
                self._current_loc[0] = self._stride[0]
                for i in range(self.iterator_support - 1):
                    self._current_loc[i + 1] = 0
            elif self._counter_update:
                for i in range(self.iterator_support):
                    if self._update[i] & (i < self._dimensionality):
                        if self._dim_counter[i] == (self._range[i] - 1):
                            self._current_loc[i] = 0
                        else:
                            self._current_loc[i] = self._current_loc[i] + self._stride[i]

    @always((posedge, "clk"), (posedge, "reset"))
    def ping_npong_update(self):
        if self._reset:
            self._ping_npong = 0
        elif self._clk_en:
            if self._flush:
                self._ping_npong = 0
            elif self._autoswitch | self._switch:
                self._ping_npong = ~self._ping_npong

    @always((posedge, "clk"), (posedge, "reset"))
    def write_addr_update(self):
        if self._reset:
            self._write_addr = 0
        elif self._clk_en:
            if self._flush:
                self._write_addr = 0
            elif self._autoswitch | self._switch:
                self._write_addr = 0
            elif self._wen & ~self._write_done & ~self._write_done_thresh:
                self._write_addr = self._write_addr + 1

    @always((posedge, "clk"), (posedge, "reset"))
    def read_cnt_update(self):
        if self._reset:
            self._read_cnt = 0
        elif self._clk_en:
            if self._flush:
                self._read_cnt = 0
            elif self._autoswitch | self._switch:
                self._read_cnt = 0
            elif self._counter_update:
                self._read_cnt = self._read_cnt + 1


if __name__ == "__main__":
    db_dut = DoubleBufferControl(16, 512, 2, 6)
    verilog(db_dut, filename="doublebuffer_control.sv", check_active_high=False)



from lake.models.model import Model


class PassthruModel(Model):

    def __init__(self):
        self._val = 0

    def set_config(self, new_config):
        # Configure top level
        for key, config_val in new_config.items():
            if key not in self.config:
                AssertionError("Gave bad config...")
            else:
                self.config[key] = config_val

    def set_in(self, in_data):
        return in_data



from kratos import *


class PassThroughMod(Generator):
    def __init__(self):
        super().__init__("PassThrough")

        self._clk = self.clock("clk")
        self._data_in = self.input("data_in", 1)
        self._data_out = self.output("data_out", 1)
        self.wire(self._data_out, self._data_in)



from kratos import *
from math import log


class SRAMStub(Generator):

    ##########################
    # Generation             #
    ##########################
    def __init__(self,
                 data_width,
                 width_mult,
                 depth):
        super().__init__("sram_stub")
        self.data_width = data_width
        self.width_mult = width_mult
        self.depth = depth

        ############################
        # Clock and Reset          #
        ############################
        self._clk = self.clock("clk")

        ############################
        # Inputs                   #
        ############################
        self._wen = self.input("wen", 1)
        self._cen = self.input("cen", 1)
        self._addr = self.input("addr", clog2(self.depth))
        # self._data_in = self.input("data_in", self.data_width)
        self._data_in = self.input("data_in",
                                   self.data_width,
                                   size=self.width_mult,
                                   explicit_array=True,
                                   packed=True)

        ############################
        # Outputs                  #
        ############################
        # self._data_out = self.output("data_out", self.data_width)
        self._data_out = self.output("data_out",
                                     self.data_width,
                                     size=self.width_mult,
                                     explicit_array=True,
                                     packed=True)

        ############################
        # Local Variables          #
        ############################
        # self._data_array = self.var("data_array",
        #                             self.data_width,
        #                             size=self.depth,
        #                             packed=True,
        #                             explicit_array=True)
        self._data_array = self.var("data_array",
                                    self.data_width,
                                    size=(self.depth,
                                          self.width_mult),
                                    packed=True,
                                    explicit_array=True)

        ############################
        # Add seq blocks           #
        ############################
        self.add_code(self.seq_data_access)
        self.add_code(self.seq_data_out)

    ##########################
    # Access sram array      #
    ##########################
    @always_ff((posedge, "clk"))
    def seq_data_access(self):
        if self._cen & self._wen:
            self._data_array[self._addr] = self._data_in

    @always_ff((posedge, "clk"))
    def seq_data_out(self):
        if self._cen & ~self._wen:
            self._data_out = self._data_array[self._addr]


if __name__ == "__main__":
    dut = SRAMStub(16, 1, 1024)
    verilog(dut, filename="sram_stub.sv")



from lake.models.model import Model
import kratos as kts


class SRAMModel(Model):
    def __init__(self,
                 data_width,
                 width_mult,
                 depth,
                 num_tiles):
        self.data_width = data_width
        self.width_mult = width_mult
        self.depth = depth
        self.num_tiles = num_tiles
        self.address_width = kts.clog2(self.num_tiles * self.depth)

        self.chain_idx_bits = max(1, kts.clog2(num_tiles))

        self.chain_idx_tile = 0

        self.rd_reg = []
        for i in range(self.width_mult):
            self.rd_reg.append(0)
        self.mem = []
        for i in range(self.depth):
            row = []
            for j in range(self.width_mult):
                row.append(0)
            self.mem.append(row)

    def set_config(self, new_config):
        # No configuration space
        return

    def interact(self,
                 wen,
                 cen,
                 addr,
                 data):
        '''
        Returns (rd_reg)
        '''

        rd_reg_ret = self.rd_reg

        addr = addr % self.depth

        # no-op
        if cen == 0:
            return rd_reg_ret
        elif wen == 1:
            self.mem[addr] = data.copy()
        else:
            # Read
            self.rd_reg = self.mem[addr]

        return list(rd_reg_ret)

    def get_rd_reg(self):
        return list(self.rd_reg)

    def dump_mem(self):
        for i in range(self.depth):
            print(f"addr: {i}, data: {self.mem[i]}")



from kratos import *
import kratos as kts


class Chain(Generator):
    def __init__(self,
                 data_width,
                 interconnect_output_ports,
                 chain_idx_bits):
        super().__init__("Chain", debug=True)

        # generator parameters
        self.data_width = data_width
        self.interconnect_output_ports = interconnect_output_ports
        self.chain_idx_bits = chain_idx_bits

        # inputs

        self._enable_chain_output = self.input("enable_chain_output", 1)

        self._chain_idx_output = self.input("chain_idx_output",
                                            self.chain_idx_bits)

        self._curr_tile_valid_out = self.input("curr_tile_valid_out",
                                               self.interconnect_output_ports)

        self._curr_tile_data_out = self.input("curr_tile_data_out",
                                              self.data_width,
                                              size=self.interconnect_output_ports,
                                              packed=True,
                                              explicit_array=True)

        self._chain_valid_in = self.input("chain_valid_in",
                                          self.interconnect_output_ports)

        self._chain_data_in = self.input("chain_data_in",
                                         self.data_width,
                                         size=self.interconnect_output_ports,
                                         packed=True,
                                         explicit_array=True)

        self._chain_data_out = self.output("chain_data_out",
                                           self.data_width,
                                           size=self.interconnect_output_ports,
                                           packed=True,
                                           explicit_array=True)

        self._chain_valid_out = self.output("chain_valid_out",
                                            self.interconnect_output_ports)

        self._chain_data_out_inter = self.var("chain_data_out_inter",
                                              self.data_width,
                                              size=self.interconnect_output_ports,
                                              packed=True,
                                              explicit_array=True)

        self._chain_valid_out_inter = self.var("chain_valid_out_inter",
                                               self.interconnect_output_ports)

        self._data_out_tile = self.output("data_out_tile",
                                          self.data_width,
                                          size=self.interconnect_output_ports,
                                          packed=True,
                                          explicit_array=True)

        self._valid_out_tile = self.output("valid_out_tile",
                                           self.interconnect_output_ports)

        self.add_code(self.set_data_out)
        self.add_code(self.set_valid_out)
        self.add_code(self.set_chain_data)
        self.add_code(self.set_chain_valid)
        self.add_code(self.set_chain_values)

    @always_comb
    def set_data_out(self):
        if self._enable_chain_output:
            self._data_out_tile = self._chain_data_out_inter
        else:
            self._data_out_tile = self._curr_tile_data_out

    @always_comb
    def set_valid_out(self):
        if self._enable_chain_output:
            if ~(self._chain_idx_output == 0):
                for i in range(self.interconnect_output_ports):
                    self._valid_out_tile[i] = 0
            else:
                self._valid_out_tile = self._chain_valid_out_inter
        else:
            for j in range(self.interconnect_output_ports):
                self._valid_out_tile[j] = self._curr_tile_valid_out[j]

    @always_comb
    def set_chain_data(self):
        self._chain_data_out = self._chain_data_out_inter

    @always_comb
    def set_chain_valid(self):
        if (self._chain_idx_output == 0) | (~self._enable_chain_output):
            for i in range(self.interconnect_output_ports):
                self._chain_valid_out[i] = 0
        else:
            self._chain_valid_out = self._chain_valid_out_inter

    @always_comb
    def set_chain_values(self):
        for i in range(self.interconnect_output_ports):
            if self._chain_valid_in[i] == 0:
                self._chain_data_out_inter[i] = self._curr_tile_data_out[i]
                self._chain_valid_out_inter[i] = self._curr_tile_valid_out[i]
            else:
                self._chain_data_out_inter[i] = self._chain_data_in[i]
                self._chain_valid_out_inter[i] = self._chain_valid_in[i]


if __name__ == "__main__":
    dut = Chain(data_width=16,
                interconnect_output_ports=3,
                chain_idx_bits=1)
    verilog(dut, filename="chain.sv")



import kratos
from kratos import *
from math import log
from lake.modules.transpose_buffer import TransposeBuffer
from lake.passes.passes import lift_config_reg


class TransposeBufferAggregation(Generator):
    def __init__(self,
                 # number of bits in a word
                 word_width,
                 # number of words that can be sotred at an address in SRAM
                 # note fetch_width must be powers of 2
                 fetch_width,
                 # total number of transpose buffers
                 num_tb,
                 # height of this particular transpose buffer
                 max_tb_height,
                 # maximum value for range parameters in nested for loop
                 # (and as a result, maximum length of indices input vector)
                 # specifying inner for loop values for output column
                 # addressing
                 max_range,
                 max_range_inner,
                 max_stride,
                 tb_iterator_support):
        super().__init__("transpose_buffer_aggregation")

        # generation parameters
        self.word_width = word_width
        self.fetch_width = fetch_width
        self.num_tb = num_tb
        self.max_tb_height = max_tb_height
        self.max_range = max_range
        self.max_range_inner = max_range_inner
        self.max_stride = max_stride
        self.tb_iterator_support = tb_iterator_support

        self.num_tb_bits = max(1, clog2(self.num_tb))

        # inputs
        self.clk = self.clock("clk")
        # active low asynchronous reset
        self.rst_n = self.reset("rst_n", 1)
        self.SRAM_to_tb_data = self.input("SRAM_to_tb_data",
                                          width=self.word_width,
                                          size=self.fetch_width,
                                          packed=True)

        self.valid_data = self.input("valid_data", 1)
        self.tb_index_for_data = self.input("tb_index_for_data", self.num_tb_bits)

        self.tba_ren = self.input("tba_ren", 1)

        # Ack the ready
        self._ack_in = self.input("ack_in", 1)

        self.mem_valid_data = self.input("mem_valid_data", 1)

        # outputs
        self.tb_to_interconnect_data = self.output("tb_to_interconnect_data",
                                                   width=self.word_width,
                                                   size=self.max_tb_height,
                                                   packed=True)
        self.tb_to_interconnect_valid = self.output("tb_to_interconnect_valid", 1)
        self.tb_arbiter_rdy = self.output("tb_arbiter_rdy", 1)

        # local variables
        self.valid_data_all = self.var("valid_data_all",
                                       width=1,
                                       size=self.num_tb,
                                       packed=True)

        self.tb_output_data_all = self.var("tb_output_data_all",
                                           width=self.word_width,
                                           size=[self.num_tb, self.max_tb_height],
                                           packed=True)
        self.tb_output_valid_all = self.var("tb_output_valid_all", width=1, size=self.num_tb, packed=True)
        self.tb_arbiter_rdy_all = self.var("tb_arbiter_rdy_all", width=1, size=self.num_tb, packed=True)
        self.output_valid = self.var("output_valid", 1)
        self.output_inter = self.var("output_inter",
                                     width=self.word_width,
                                     size=self.max_tb_height,
                                     packed=True)

        for i in range(self.num_tb):
            self.add_child(f"tb_{i}",
                           TransposeBuffer(self.word_width,
                                           self.fetch_width,
                                           self.num_tb,
                                           self.max_tb_height,
                                           self.max_range,
                                           self.max_range_inner,
                                           self.max_stride,
                                           self.tb_iterator_support),
                           clk=self.clk,
                           rst_n=self.rst_n,
                           input_data=self.SRAM_to_tb_data,
                           valid_data=self.valid_data_all[i],
                           col_pixels=self.tb_output_data_all[i],
                           output_valid=self.tb_output_valid_all[i],
                           rdy_to_arbiter=self.tb_arbiter_rdy_all[i],
                           ack_in=self._ack_in,
                           mem_valid_data=self.mem_valid_data,
                           ren=self.tba_ren)

        self.add_code(self.set_valid_data_all)
        self.set_output_valid()
        self.send_tba_rdy()

    @always_comb
    def set_valid_data_all(self):
        if ~self.valid_data:
            for i in range(self.num_tb):
                self.valid_data_all[i] = 0
        else:
            for i in range(self.num_tb):
                if i == self.tb_index_for_data:
                    self.valid_data_all[i] = 1
                else:
                    self.valid_data_all[i] = 0

    def set_output_valid(self):
        comb_output = self.combinational()
        valid_count = self.tb_output_valid_all[0]
        comb_output.add_stmt(self.output_inter.assign(self.tb_output_data_all[0]))
        for i in range(1, self.num_tb):
            valid_count = valid_count + self.tb_output_valid_all[i]
            if_vld = IfStmt(self.tb_output_valid_all[i] == 1)
            if_vld.then_(self.output_inter.assign(self.tb_output_data_all[i]))
            comb_output.add_stmt(if_vld)
        if_vld_count = IfStmt(valid_count > 0)
        if_vld_count.then_(self.tb_to_interconnect_valid.assign(1))
        if_vld_count.else_(self.tb_to_interconnect_valid.assign(0))
        comb_output.add_stmt(self.tb_to_interconnect_data.assign(self.output_inter))
        comb_output.add_stmt(if_vld_count)

    def send_tba_rdy(self):
        comb_tb_arbiter_rdy = self.combinational()
        rdy_count = self.tb_arbiter_rdy_all[0]
        for i in range(1, self.num_tb):
            rdy_count = rdy_count + self.tb_arbiter_rdy_all[i]
        if_rdy_count = IfStmt(rdy_count > 0)
        if_rdy_count.then_(self.tb_arbiter_rdy.assign(1))
        if_rdy_count.else_(self.tb_arbiter_rdy.assign(0))
        comb_tb_arbiter_rdy.add_stmt(if_rdy_count)


if __name__ == "__main__":
    dut = TransposeBufferAggregation(word_width=1,
                                     fetch_width=4,
                                     num_tb=3,
                                     max_tb_height=1,
                                     max_range=5,
                                     max_range_inner=3,
                                     max_stride=2,
                                     tb_iterator_support=2)
    verilog(dut, filename="tba.sv", additional_passes={"lift config regs": lift_config_reg})



from kratos import *
from lake.utils.util import increment, decrement
from lake.attributes.config_reg_attr import ConfigRegAttr


class AggAligner(Generator):
    '''
    This module will help with the word alignment of the aggregator. This module
    will keep a running count of how many aggregators have been filled based
    on the line length (or some other notion of termination)
    '''
    def __init__(self,
                 data_width,
                 max_line_length):
        super().__init__("agg_aligner")

        # Capture to the object
        self.data_width = data_width
        self.max_line_length = max_line_length
        self.counter_width = clog2(self.max_line_length)

        # Clock and Reset
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        # Inputs
        self._in_dat = self.input("in_dat", self.data_width)
        self._in_valid = self.input("in_valid", 1)

        self._line_length = self.input("line_length", self.counter_width)
        self._line_length.add_attribute(ConfigRegAttr("Line Length/Image Width for alignment"))

        # Outputs
        self._out_dat = self.output("out_dat", self.data_width)
        self._out_valid = self.output("out_valid", 1)
        self._out_align = self.output("align", 1)

        # Local Signals
        self._cnt = self.var("cnt", self.counter_width)

        # Generate
        self.add_code(self.update_cnt)
        self.add_code(self.set_align)

        self.wire(self._out_dat, self._in_dat)
        self.wire(self._out_valid, self._in_valid)

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def update_cnt(self):
        if ~self._rst_n:
            self._cnt = 0
        elif(self._in_valid):
            if(self._cnt == decrement(self._line_length, 1)):
                self._cnt = 0
            else:
                self._cnt = increment(self._cnt, 1)

    @always_comb
    def set_align(self):
        self._out_align = self._in_valid & (self._cnt == decrement(self._line_length, 1))


if __name__ == "__main__":
    align_dut = AggAligner(data_width=16,
                           max_line_length=64)
    verilog(align_dut, filename="agg_aligner.sv")



from kratos import *
from lake.modules.passthru import *
from lake.modules.register_file import RegisterFile
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.attributes.range_group import RangeGroupAttr
from lake.passes.passes import lift_config_reg
from lake.modules.sram_stub import SRAMStub
from lake.modules.for_loop import ForLoop
from lake.modules.addr_gen import AddrGen
from lake.modules.spec.sched_gen import SchedGen
import kratos as kts


class StrgUB(Generator):
    def __init__(self,
                 data_width=16,  # CGRA Params
                 mem_width=64,
                 mem_depth=512,
                 banks=1,
                 input_addr_iterator_support=6,
                 output_addr_iterator_support=6,
                 input_sched_iterator_support=6,
                 output_sched_iterator_support=6,
                 config_width=16,
                 #  output_config_width=16,
                 interconnect_input_ports=2,  # Connection to int
                 interconnect_output_ports=2,
                 mem_input_ports=1,
                 mem_output_ports=1,
                 read_delay=1,  # Cycle delay in read (SRAM vs Register File)
                 rw_same_cycle=False,  # Does the memory allow r+w in same cycle?
                 agg_height=4,
                 max_line_length=128,
                 max_tb_height=1,
                 tb_range_inner_max=5,
                 tb_sched_max=64,
                 num_tiles=1):
        super().__init__("strg_ub", debug=True)

        self.fetch_width = mem_width // data_width
        self.interconnect_input_ports = interconnect_input_ports
        self.interconnect_output_ports = interconnect_output_ports
        self.agg_height = agg_height
        self.mem_depth = mem_depth
        # generation parameters
        # inputs
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        self._clk_en = self.input("clk_en", 1)
        self._flush = self.reset("flush", is_async=False, active_high=True)

        self._data_in = self.input("data_in", data_width,
                                   size=self.interconnect_input_ports,
                                   packed=True,
                                   explicit_array=True)

        # Create cycle counter to share...
        self._cycle_count = self.var("cycle_count", 16)
        self.add_code(self.increment_cycle_count)

        # outputs
        self._data_out = self.output("data_out", data_width,
                                     size=self.interconnect_output_ports,
                                     packed=True,
                                     explicit_array=True)

        # local variables
        self._write = self.var("write", 1)
        self._read = self.var("read", 1)
        self._read_d1 = self.var("read_d1", 1)
        self.add_code(self.delay_read)

        self._write_addr = self.var("write_addr", config_width)
        self._read_addr = self.var("read_addr", config_width)
        self._addr = self.var("addr", clog2(mem_depth))

        self._agg_write = self.var("agg_write", self.interconnect_input_ports)

        # Make this based on the size
        self._agg_write_addr = self.var("agg_write_addr", 2 + clog2(self.agg_height),
                                        size=self.interconnect_input_ports,
                                        packed=True,
                                        explicit_array=True)
        self._agg_read_addr = self.var("agg_read_addr", max(1, clog2(self.agg_height)),
                                       size=self.interconnect_input_ports,
                                       packed=True,
                                       explicit_array=True)

        self.agg_rd_addr_gen_width = 8
        self._agg_read_addr_gen_out = self.var("agg_read_addr_gen_out", self.agg_rd_addr_gen_width,
                                               size=self.interconnect_input_ports,
                                               packed=True,
                                               explicit_array=True)

        self._sram_write_data = self.var("sram_write_data", data_width,
                                         size=self.fetch_width,
                                         packed=True)
        self._sram_read_data = self.var("sram_read_data", data_width,
                                        size=self.fetch_width,
                                        packed=True,
                                        explicit_array=True)

        self._data_to_sram = self.output("data_to_strg", data_width,
                                         size=self.fetch_width,
                                         packed=True)
        self._data_from_sram = self.input("data_from_strg", data_width,
                                          size=self.fetch_width,
                                          packed=True)

        self._wen_to_sram = self.output("wen_to_strg", 1, packed=True)
        self._cen_to_sram = self.output("cen_to_strg", 1, packed=True)

        self._addr_to_sram = self.output("addr_out", clog2(mem_depth), packed=True)

        self.wire(self._addr_to_sram, self._addr)
        self.wire(self._data_to_sram, self._sram_write_data)
        self.wire(self._data_from_sram, self._sram_read_data)
        self.wire(self._wen_to_sram, self._write)
        self.wire(self._cen_to_sram, self._write | self._read)

        self._agg_write_index = self.var("agg_write_index", 2, size=4)

        self._output_port_sel_addr = self.var("output_port_sel_addr",
                                              max(1, clog2(self.interconnect_output_ports)))

        self.agg_write_scheds = []
        self.agg_read_addrs = []
        self._input_port_sel_addr = self.var("input_port_sel_addr",
                                             max(1, clog2(self.interconnect_input_ports)))
        # Create an input to agg write scheduler + addressor for each input
        # Also need an addressor for the mux in addition to the read addr
        self._agg = self.var(f"agg",
                             width=data_width,
                             size=(self.interconnect_input_ports,
                                   self.agg_height,
                                   self.fetch_width),
                             packed=True,
                             explicit_array=True)

        for i in range(self.interconnect_input_ports):

            forloop_ctr = ForLoop(iterator_support=4,
                                  config_width=self._agg_write_addr.width)
            loop_itr = forloop_ctr.get_iter()
            loop_wth = forloop_ctr.get_cfg_width()

            self.add_child(f"loops_in2buf_{i}",
                           forloop_ctr,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=self._agg_write[i])

            newAG = AddrGen(iterator_support=loop_itr,
                            config_width=loop_wth)
            self.add_child(f"agg_write_addr_gen_{i}",
                           newAG,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=self._agg_write[i],
                           mux_sel=forloop_ctr.ports.mux_sel_out,
                           addr_out=self._agg_write_addr[i])

            newSG = SchedGen(iterator_support=loop_itr,
                             config_width=loop_wth)
            self.agg_write_scheds.append(newSG)
            self.add_child(f"agg_write_sched_gen_{i}",
                           newSG,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           mux_sel=forloop_ctr.ports.mux_sel_out,
                           cycle_count=self._cycle_count,
                           valid_output=self._agg_write[i])

            forloop_ctr_rd = ForLoop(iterator_support=4,
                                     config_width=self.agg_rd_addr_gen_width)
            loop_itr = forloop_ctr_rd.get_iter()
            loop_wth = forloop_ctr_rd.get_cfg_width()

            self.add_child(f"loops_in2buf_autovec_read_{i}",
                           forloop_ctr_rd,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=(self._write &
                                 (self._input_port_sel_addr == const(i, self._input_port_sel_addr.width))))

            newAG = AddrGen(iterator_support=loop_itr,
                            config_width=loop_wth)
            self.agg_read_addrs.append(newAG)
            self.add_child(f"agg_read_addr_gen_{i}",
                           newAG,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=(self._write &
                                 (self._input_port_sel_addr == const(i, self._input_port_sel_addr.width))),
                           mux_sel=forloop_ctr_rd.ports.mux_sel_out,
                           addr_out=self._agg_read_addr_gen_out[i])
            self.wire(self._agg_read_addr[i], self._agg_read_addr_gen_out[i][self._agg_read_addr.width - 1, 0])

        # Create for loop counters that can be shared across the input port selection and SRAM write
        fl_ctr_sram_wr = ForLoop(iterator_support=6,
                                 config_width=16)
        loop_itr = fl_ctr_sram_wr.get_iter()
        loop_wth = fl_ctr_sram_wr.get_cfg_width()

        self.add_child(f"loops_in2buf_autovec_write",
                       fl_ctr_sram_wr,
                       clk=self._clk,
                       rst_n=self._rst_n,
                       step=self._write)

        # Now we determine what data goes through to the sram...
        # If we have more than one port, we can generate a selector
        # to pick which input port should go through - then we send
        # the step signal to the appropriate input port
        if self.interconnect_input_ports > 1:
            self.add_child(f"port_sel_addr",
                           AddrGen(iterator_support=loop_itr,
                                   config_width=clog2(self.interconnect_input_ports)),
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=self._write,
                           mux_sel=fl_ctr_sram_wr.ports.mux_sel_out,
                           addr_out=self._input_port_sel_addr)
            # Addr for port select should be driven on agg to sram write sched
        else:
            self.wire(self._input_port_sel_addr[0], const(0, self._input_port_sel_addr.width))

        # Whatever comes through here should hopefully just pipe through seamlessly
        # addressor modules
        self.add_child(f"input_addr_gen",
                       AddrGen(input_addr_iterator_support,
                               config_width),
                       clk=self._clk,
                       rst_n=self._rst_n,
                       step=self._write,
                       mux_sel=fl_ctr_sram_wr.ports.mux_sel_out,
                       addr_out=self._write_addr)

        # scheduler modules
        self.add_child(f"input_sched_gen",
                       SchedGen(input_sched_iterator_support,
                                config_width),
                       clk=self._clk,
                       rst_n=self._rst_n,
                       cycle_count=self._cycle_count,
                       mux_sel=fl_ctr_sram_wr.ports.mux_sel_out,
                       valid_output=self._write)

        # -------------------------------- Delineate new group -------------------------------
        fl_ctr_sram_rd = ForLoop(iterator_support=6,
                                 config_width=16)
        loop_itr = fl_ctr_sram_rd.get_iter()
        loop_wth = fl_ctr_sram_rd.get_cfg_width()

        self.add_child(f"loops_buf2out_autovec_read",
                       fl_ctr_sram_rd,
                       clk=self._clk,
                       rst_n=self._rst_n,
                       step=self._read)

        self.add_child(f"output_addr_gen",
                       AddrGen(iterator_support=6,
                               config_width=16),
                       clk=self._clk,
                       rst_n=self._rst_n,
                       step=self._read,
                       mux_sel=fl_ctr_sram_rd.ports.mux_sel_out,
                       addr_out=self._read_addr)

        self.add_child(f"output_sched_gen",
                       SchedGen(iterator_support=6,
                                config_width=16),
                       clk=self._clk,
                       rst_n=self._rst_n,
                       cycle_count=self._cycle_count,
                       mux_sel=fl_ctr_sram_rd.ports.mux_sel_out,
                       valid_output=self._read)

        self._tb_read = self.var("tb_read", self.interconnect_output_ports)
        self.tb_height = 4

        self._tb_write_addr = self.var("tb_write_addr", 6,
                                       size=self.interconnect_output_ports,
                                       packed=True,
                                       explicit_array=True)
        self._tb_read_addr = self.var("tb_read_addr", 6,
                                      size=self.interconnect_output_ports,
                                      packed=True,
                                      explicit_array=True)

        self._tb = self.var("tb",
                            width=data_width,
                            size=(self.interconnect_output_ports,
                                  self.tb_height,
                                  self.fetch_width),
                            packed=True,
                            explicit_array=True)

        for i in range(self.interconnect_output_ports):
            fl_ctr_tb_wr = ForLoop(iterator_support=2,
                                   config_width=6)
            loop_itr = fl_ctr_tb_wr.get_iter()
            loop_wth = fl_ctr_tb_wr.get_cfg_width()

            self.add_child(f"loops_buf2out_autovec_write_{i}",
                           fl_ctr_tb_wr,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=self._read_d1 & (self._output_port_sel_addr ==
                                                 const(i, self._output_port_sel_addr.width)))

            self.add_child(f"tb_write_addr_gen_{i}",
                           AddrGen(iterator_support=loop_itr,
                                   config_width=loop_wth),
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=self._read_d1 & (self._output_port_sel_addr ==
                                                 const(i, self._output_port_sel_addr.width)),
                           mux_sel=fl_ctr_tb_wr.ports.mux_sel_out,
                           addr_out=self._tb_write_addr[i])

            fl_ctr_tb_rd = ForLoop(iterator_support=2,
                                   config_width=16)
            loop_itr = fl_ctr_tb_rd.get_iter()
            loop_wth = fl_ctr_tb_rd.get_cfg_width()

            self.add_child(f"loops_buf2out_read_{i}",
                           fl_ctr_tb_rd,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=self._tb_read[i])

            self.add_child(f"tb_read_addr_gen_{i}",
                           AddrGen(iterator_support=loop_itr,
                                   config_width=6),
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=self._tb_read[i],
                           mux_sel=fl_ctr_tb_rd.ports.mux_sel_out,
                           addr_out=self._tb_read_addr[i])

            self.add_child(f"tb_read_sched_gen_{i}",
                           SchedGen(iterator_support=loop_itr,
                                    config_width=16),
                           clk=self._clk,
                           rst_n=self._rst_n,
                           cycle_count=self._cycle_count,
                           mux_sel=fl_ctr_tb_rd.ports.mux_sel_out,
                           valid_output=self._tb_read[i])

        if self.interconnect_output_ports > 1:

            fl_ctr_out_sel = ForLoop(iterator_support=2,
                                     config_width=clog2(self.interconnect_output_ports))
            loop_itr = fl_ctr_out_sel.get_iter()
            loop_wth = fl_ctr_out_sel.get_cfg_width()

            self.add_child(f"loops_buf2out_out_sel",
                           fl_ctr_out_sel,
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=self._read_d1)

            self.add_child(f"out_port_sel_addr",
                           AddrGen(iterator_support=loop_itr,
                                   config_width=loop_wth),
                           clk=self._clk,
                           rst_n=self._rst_n,
                           step=self._read_d1,
                           mux_sel=fl_ctr_out_sel.ports.mux_sel_out,
                           addr_out=self._output_port_sel_addr)
            # Addr for port select should be driven on agg to sram write sched
        else:
            self.wire(self._output_port_sel_addr[0], const(0, self._output_port_sel_addr.width))

        # lift_config_reg(self.internal_generator)

        self.add_code(self.set_sram_addr)
        for idx in range(self.interconnect_input_ports):
            self.add_code(self.agg_ctrl, idx=idx)

        self.add_code(self.agg_to_sram)
        self.add_code(self.tb_ctrl)

        for idx in range(self.interconnect_output_ports):
            self.add_code(self.tb_to_out, idx=idx)

    @always_comb
    def set_sram_addr(self):
        if self._write:
            self._addr = self._write_addr[clog2(self.mem_depth) - 1, 0]
        else:
            self._addr = self._read_addr[clog2(self.mem_depth) - 1, 0]

    @always_ff((posedge, "clk"))
    def agg_ctrl(self, idx):
        if self._agg_write[idx]:
            self._agg[idx][self._agg_write_addr[idx][self._agg_write_addr[0].width - 1, 2]]\
                          [self._agg_write_addr[idx][1, 0]] = self._data_in[idx]

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def delay_read(self):
        if ~self._rst_n:
            self._read_d1 = 0
        else:
            self._read_d1 = self._read

    @always_comb
    def agg_to_sram(self):
        for i in range(self.fetch_width):
            self._sram_write_data[i] = \
                self._agg[self._input_port_sel_addr][self._agg_read_addr[self._input_port_sel_addr]][i]

    @always_ff((posedge, "clk"))
    def tb_ctrl(self):
        if self._read_d1:
            self._tb[self._output_port_sel_addr][self._tb_write_addr[self._output_port_sel_addr][1, 0]] = \
                self._sram_read_data

    @always_comb
    def tb_to_out(self, idx):
        self._data_out[idx] = self._tb[idx][self._tb_read_addr[idx][3, 2]][self._tb_read_addr[idx][1, 0]]

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def increment_cycle_count(self):
        if ~self._rst_n:
            self._cycle_count = 0
        else:
            self._cycle_count = self._cycle_count + 1


if __name__ == "__main__":
    lake_dut = StrgUB()
    verilog(lake_dut, filename="strg_ub_new.sv",
            optimize_if=False,
            additional_passes={"lift config regs": lift_config_reg})



from kratos import *
from lake.modules.aggregator import Aggregator
import kratos as kts
from lake.attributes.config_reg_attr import ConfigRegAttr


class AggregationBuffer(Generator):
    '''
    Aggregation buffer handles an item coming in and
    directs it to the proper row based on the scheduling
    '''
    def __init__(self,
                 agg_height,
                 data_width,
                 mem_width,
                 max_agg_schedule):

        super().__init__("aggregation_buffer")

        self.agg_height = agg_height
        self.data_width = data_width
        self.mem_width = mem_width
        self.max_agg_schedule = max_agg_schedule  # This is the maximum length of the schedule
        self.fw_int = int(self.mem_width / self.data_width)

        # Clock and Reset
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        # Inputs
        # Bring in a single element into an AggregationBuffer w/ valid signaling
        self._data_in = self.input("data_in", self.data_width)
        self._valid_in = self.input("valid_in", 1)
        self._align = self.input("align", 1)
        # Outputs
        self._data_out = self.output("data_out", self.mem_width)
        self._valid_out = self.output("valid_out", 1)

        self._data_out_chop = []
        for i in range(self.fw_int):
            self._data_out_chop.append(self.output(f"data_out_chop_{i}", self.data_width))
            self.add_stmt(
                self._data_out_chop[i].assign(self._data_out[(self.data_width * (i + 1)) - 1,
                                                             self.data_width * i]))

        # CONFIG:
        # We receive a periodic (doesn't need to be, but has a maximum schedule,
        # so...possibly the schedule is a for loop?
        # Tells us where to write successive elements...
        self._in_schedule = self.input("in_sched",
                                       max(1, clog2(self.agg_height)),
                                       size=self.max_agg_schedule,
                                       explicit_array=True,
                                       packed=True)
        doc = "Input schedule for aggregation buffer. Enumerate which" + \
              f" of {self.agg_height} buffers to write to."
        self._in_schedule.add_attribute(ConfigRegAttr(doc))

        self._in_period = self.input("in_period", clog2(self.max_agg_schedule))
        doc = "Input period for aggregation buffer. 1 is a reasonable" + \
              " setting for most applications"
        self._in_period.add_attribute(ConfigRegAttr(doc))

        # ...and which order to output the blocks
        self._out_schedule = self.input("out_sched",
                                        max(1, clog2(agg_height)),
                                        size=self.max_agg_schedule,
                                        explicit_array=True,
                                        packed=True)
        doc = "Output schedule for aggregation buffer. Enumerate which" + \
              f" of {self.agg_height} buffers to write to SRAM from."
        self._out_schedule.add_attribute(ConfigRegAttr(doc))

        self._out_period = self.input("out_period", clog2(self.max_agg_schedule))
        self._out_period.add_attribute(ConfigRegAttr("Output period for aggregation buffer"))

        self._in_sched_ptr = self.var("in_sched_ptr", clog2(self.max_agg_schedule))
        self._out_sched_ptr = self.var("out_sched_ptr", clog2(self.max_agg_schedule))

        # Local Signals
        self._aggs_out = self.var("aggs_out",
                                  self.mem_width,
                                  size=self.agg_height,
                                  packed=True,
                                  explicit_array=True)

        self._aggs_sep = []
        for i in range(self.agg_height):
            self._aggs_sep.append(self.var(f"aggs_sep_{i}",
                                           self.data_width,
                                           size=self.fw_int,
                                           packed=True))
        self._valid_demux = self.var("valid_demux", self.agg_height)
        self._align_demux = self.var("align_demux", self.agg_height)
        self._next_full = self.var("next_full", self.agg_height)
        self._valid_out_mux = self.var("valid_out_mux", self.agg_height)
        for i in range(self.agg_height):
            # Add in the children aggregators...
            self.add_child(f"agg_{i}",
                           Aggregator(self.data_width,
                                      mem_word_width=self.fw_int),
                           clk=self._clk,
                           rst_n=self._rst_n,
                           in_pixels=self._data_in,
                           valid_in=self._valid_demux[i],
                           agg_out=self._aggs_sep[i],
                           valid_out=self._valid_out_mux[i],
                           next_full=self._next_full[i],
                           align=self._align_demux[i])
            portlist = []
            if self.fw_int == 1:
                self.wire(self._aggs_out[i], self._aggs_sep[i])
            else:
                for j in range(self.fw_int):
                    portlist.append(self._aggs_sep[i][self.fw_int - 1 - j])
                self.wire(self._aggs_out[i], kts.concat(*portlist))

        # Sequential code blocks
        self.add_code(self.update_in_sched_ptr)
        self.add_code(self.update_out_sched_ptr)

        # Combinational code blocks
        self.add_code(self.valid_demux_comb)
        self.add_code(self.align_demux_comb)
        self.add_code(self.valid_out_comb)
        self.add_code(self.output_data_comb)

    # Update the pointer and mux for input and output schedule
    # Now, obey the input schedule to send to the proper Aggregator
    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def update_in_sched_ptr(self):
        if ~self._rst_n:
            self._in_sched_ptr = 0
        elif self._next_full[self._in_schedule[self._in_sched_ptr]]:
            self._in_sched_ptr = \
                ternary(self._in_sched_ptr == (self._in_period - 1),
                        const(0, self._in_sched_ptr.width),
                        self._in_sched_ptr + const(1, self._in_sched_ptr.width))

    @always_ff((posedge, "clk"), (negedge, "rst_n"))
    def update_out_sched_ptr(self):
        if ~self._rst_n:
            self._out_sched_ptr = 0
        elif self._valid_out:
            self._out_sched_ptr = \
                ternary(self._out_sched_ptr == (self._out_period - 1),
                        const(0, self._out_sched_ptr.width),
                        self._out_sched_ptr + const(1, self._out_sched_ptr.width))

    @always_comb
    def valid_demux_comb(self):
        self._valid_demux = 0
        self._valid_demux[self._in_schedule[self._in_sched_ptr]] = self._valid_in

    @always_comb
    def align_demux_comb(self):
        self._align_demux = 0
        self._align_demux[self._in_schedule[self._in_sched_ptr]] = self._align

    @always_comb
    def valid_out_comb(self):
        self._valid_out = self._valid_out_mux[self._out_schedule[self._out_sched_ptr]]

    @always_comb
    def output_data_comb(self):
        self._data_out = self._aggs_out[self._out_schedule[self._out_sched_ptr]]


if __name__ == "__main__":
    db_dut = AggregationBuffer(agg_height=4,
                               data_width=16,
                               mem_width=64,
                               max_agg_schedule=64)

    verilog(db_dut, filename="aggregation_buffer.sv")



from lake.models.model import Model
import numpy as np


class AppCtrlModel(Model):
    '''
    Model for app ctrl
    '''
    def __init__(self,
                 int_in_ports,
                 int_out_ports,
                 sprt_stcl_valid,
                 stcl_iter_support):

        self.int_in_ports = int_in_ports
        self.int_out_ports = int_out_ports
        self.sprt_stcl_valid = sprt_stcl_valid
        self.stcl_iter_support = stcl_iter_support

        self.config = {}
        for i in range(self.int_in_ports):
            self.config[f'write_depth_{i}'] = 0
        for i in range(self.int_out_ports):
            self.config[f'read_depth_{i}'] = 0
            self.config[f'input_port_{i}'] = 0
            self.config[f'prefill_{i}'] = 0

        for i in range(self.stcl_iter_support):
            self.config[f'ranges_{i}'] = 0
            self.config[f'threshold_{i}'] = 0

        self.write_count = [0] * self.int_in_ports
        self.write_done_d = [0] * self.int_in_ports
        self.read_count = [0] * self.int_out_ports
        self.init_state = [0] * self.int_out_ports
        self.read_done_d = [0] * self.int_out_ports

    # Doesn't actually have configuration state
    def set_config(self, new_config):
        # No configuration space
        # Configure top level
        for key, config_val in new_config.items():
            if key not in self.config:
                AssertionError("Gave bad config...")
            else:
                self.config[key] = config_val

    def interact(self, wen_in, ren_in, tb_valid, ren_update):
        '''
        Returns (wen_out, ren_out, valid_out, valid_out_stencil)
        '''

        valid_out_data = tb_valid.copy()
        valid_out_stencil = tb_valid.copy()

        if self.sprt_stcl_valid:
            dim_counter = []
            update = []
            for i in range(self.stcl_iter_support):
                dim_counter.append(0)
                update.append(0)

            update[0] = 1
            for i in range(self.stcl_iter_support - 1):
                update[i + 1] = (dim_counter[i] == (self.config[f'ranges_{i}'] - 1)) and update[i]

            for i in range(self.stcl_iter_support):
                if ren_in[0] and ren_update[0]:
                    if update[i]:
                        if dim_counter[i] == (self.config[f'ranges_{i}'] - 1):
                            dim_counter[i] = 0
                        else:
                            dim_counter[i] = dim_counter[i] + 1

            threshold_comps = [dim_counter[i] >= self.config[f'threshold_{i}']
                               for i in range(self.stcl_iter_support)]

            valid_out_stencil[0] = np.bitwise_and.reduce(threshold_comps)
            for i in range(self.int_out_ports - 1):
                valid_out_stencil[i + 1] = np.bitwise_and.reduce(threshold_comps)

            for i in range(len(valid_out_data)):
                valid_out_data[i] = valid_out_data[i] & valid_out_stencil[i]

        curr_read_done_d = self.read_done_d
        curr_write_done_d = self.write_done_d

        wen_out = []
        for i in range(self.int_in_ports):
            wen_out.append((self.write_done_d[i] == 0) and (wen_in[i] == 1))

        ren_out = []
        for i in range(self.int_out_ports):
            ren_out.append((self.read_done_d[i] == 0) and
                           (ren_in[i] == 1) and
                           ((self.config[f'prefill_{i}'] == 1) or (self.init_state[i] == 1)))

        write_done = []
        for i in range(self.int_in_ports):
            wd_temp = (((wen_in[i] == 1) and (self.write_count[i] == self.config[f'write_depth_{i}'] - 1)) or
                       self.write_done_d[i] == 1)
            write_done.append(wd_temp)

        read_done = []
        for i in range(self.int_out_ports):
            rd_temp = (((ren_update[i] == 1) and
                        (ren_in[i] == 1) and
                        (self.read_count[i] == self.config[f'read_depth_{i}'] - 1)) or
                       curr_read_done_d[i] == 1 or
                       ((self.config[f'prefill_{i}'] == 0) and (self.init_state[i] == 0)))
            read_done.append(rd_temp)

        # read_done_ff
        for i in range(self.int_out_ports):
            if (write_done[self.config[f"input_port_{i}"]] == 1) and (read_done[i] == 1):
                self.read_done_d[i] = 0
            elif read_done[i] == 1:
                self.read_done_d[i] = 1

        # Deal with write done ff
        for i in range(self.int_in_ports):
            if (write_done[i] == 1) and (read_done[i] == 1):
                self.write_done_d[i] = 0
            elif write_done[i] == 1:
                self.write_done_d[i] = 1

        # Init state
        for i in range(self.int_out_ports):
            if write_done[self.config[f"input_port_{i}"]] == 1:
                self.init_state[i] = 1

        # write count
        for i in range(self.int_in_ports):
            if (write_done[i] == 1) and (read_done[i] == 1):
                self.write_count[i] = 0
            elif (wen_in[i] == 1) and (curr_write_done_d[i] == 0):
                self.write_count[i] += 1

        # read count
        for i in range(self.int_out_ports):
            if (write_done[self.config[f"input_port_{i}"]] == 1) and (read_done[i] == 1):
                self.read_count[i] = 0
            elif (ren_in[i] == 1) & (ren_update[i] == 1):
                self.read_count[i] += 1

        return (wen_out, ren_out, valid_out_data, valid_out_stencil)



from lake.models.model import Model


class PipeRegModel(Model):
    def __init__(self,
                 data_width,
                 stages):
        self.data_width = data_width
        self.stages = stages

        self.reg_stage = []
        if(self.stages > 0):
            for i in range(self.stages):
                self.reg_stage.append(0)

    # No config
    def set_config(self, new_config):
        return

    def update_data(self, data_in):
        if(self.stages == 0):
            return data_in
        else:
            data_curr = []
            for i in range(self.stages):
                data_curr.append(self.reg_stage[i])
            self.reg_stage[0] = data_in
            for i in range(self.stages - 1):
                self.reg_stage[i + 1] = data_curr[i]
            return self.reg_stage[self.stages - 1]



from kratos import *
import kratos as kts


class DemuxReads(Generator):
    '''
    Demux the SRAM reads into writes to the transpose buffers
    '''
    def __init__(self,
                 fetch_width=16,
                 data_width=16,
                 banks=1,
                 int_out_ports=2,
                 strg_rd_ports=2):

        assert not (fetch_width & (fetch_width - 1)), "Memory width needs to be a power of 2"

        super().__init__("demux_reads")
        # Absorb inputs
        self.fetch_width = fetch_width
        self.data_width = data_width
        self.fw_int = int(self.fetch_width / self.data_width)
        self.int_out_ports = int_out_ports
        self.banks = banks
        self.strg_rd_ports = strg_rd_ports

        # Clock and Reset
        self._clk = self.clock("clk")
        self._rst_n = self.reset("rst_n")

        # Inputs
        self._data_in = self.input("data_in",
                                   self.data_width,
                                   size=(self.banks * self.strg_rd_ports,
                                         self.fw_int),
                                   explicit_array=True,
                                   packed=True)

        self._valid_in = self.input("valid_in", self.banks * self.strg_rd_ports)
        self._port_in = self.input("port_in",
                                   self.int_out_ports,
                                   size=self.strg_rd_ports * self.banks,
                                   explicit_array=True,
                                   packed=True)

        # Outputs
        self._data_out = self.output("data_out",
                                     self.data_width,
                                     size=(self.int_out_ports,
                                           self.fw_int),
                                     explicit_array=True,
                                     packed=True)

        self._valid_out = self.output("valid_out",
                                      self.int_out_ports)

        self._mem_valid_data = self.input("mem_valid_data", self.banks * self.strg_rd_ports)
        self._mem_valid_data_out = self.output("mem_valid_data_out", self.int_out_ports)

        # Vars
        self._done = self.var("done", self.int_out_ports)

        self.add_code(self.set_outs)

    @always_comb
    def set_outs(self):
        for i in range(self.int_out_ports):
            self._valid_out[i] = 0
            self._data_out[i] = 0
            self._done[i] = 0
            self._mem_valid_data_out[i] = 0
            for j in range(self.banks * self.strg_rd_ports):
                if ~self._done[i]:
                    if self._valid_in[j] & self._port_in[j][i]:
                        self._valid_out[i] = 1
                        self._data_out[i] = self._data_in[j]
                        self._mem_valid_data_out[i] = self._mem_valid_data[j]
                        self._done[i] = 1


if __name__ == "__main__":
    demux_dut = DemuxReads()
    verilog(demux_dut, filename="demux_reads.sv")



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='lake-aha',
    version="0.0.3",
    author='Maxwell Strange',
    author_email='mstrange@stanford.edu',
    description='Memory Generator based on Kratos: The God of War.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/StanfordAHA/lake",
    python_requires=">=3.5",
    packages=find_packages(exclude=("tests",)),
    install_requires=[
        "kratos",
        "fault",
        "magma-lang",
        "pytest"
    ]
)



from kratos import *


class FIFOControl(Generator):
    def __init__(self, data_width, banks, mem_depth):
        super().__init__("fifo_control")

        self.data_width = data_width
        self.banks = banks
        self.mem_depth = mem_depth
        self.mem_addr_width = clog2(self.mem_depth)
        self.bank_addr_width = clog2(self.banks)
        self.full_addr = self.mem_addr_width + self.bank_addr_width

        # PORT DEFS: begin
        self._clk = self.clock("clk")
        self._clk_en = self.input("clk_en", 1)
        self._reset = self.reset("reset")
        self._flush = self.input("flush", 1)

        self._ren = self.input("ren", 1)
        self._wen = self.input("wen", 1)
        self._data_in = self.input("data_in", self.data_width)
        self._data_out = self.output("data_out", self.data_width)

        self._almost_empty = self.output("almost_empty", 1)
        self._almost_full = self.output("almost_full", 1)
        self._empty = self.output("empty", 1)
        self._full = self.output("full", 1)
        self._valid = self.output("valid", 1)

        self._fifo_to_mem_data = self.output("fifo_to_mem_data",
                                             self.data_width,
                                             size=self.banks,
                                             packed=True,
                                             explicit_array=True)
        self._fifo_to_mem_cen = self.output("fifo_to_mem_cen", self.banks)
        self._fifo_to_mem_wen = self.output("fifo_to_mem_wen", self.banks)
        self._fifo_to_mem_addr = self.output("fifo_to_mem_addr",
                                             self.mem_addr_width,
                                             size=self.banks,
                                             explicit_array=True,
                                             packed=True)
        self._mem_to_fifo_data = self.input("mem_to_fifo_data",
                                            self.data_width,
                                            size=self.banks,
                                            explicit_array=True,
                                            packed=True)

        self._num_words_mem = self.output("num_words_mem", 16)

        self._almost_count = self.input("almost_count", 4)
        self._circular_en = self.input("circular_en", 1)
        self._depth = self.input("depth", 16)
        # PORT DEFS: end

        # LOCAL VARIABLES: begin
        self._almost_count_extended = self.var("almost_count_extended", 16)
        self._read_addr = self.var("read_addr", self.full_addr)
        self._write_addr = self.var("write_addr", self.full_addr)

        self._read_addr_mem = self.var("read_addr_mem", self.mem_addr_width)
        self._write_addr_mem = self.var("write_addr_mem", self.mem_addr_width)

        self._ren_mem = self.var("ren_mem", self.banks)
        self._wen_mem = self.var("wen_mem", self.banks)
        self._cen_mem = self.var("cen_mem", self.banks)

        self._write_buffed = self.var("write_buffed", self.banks)
        self._write_buff = self.var("write_buff",
                                    self.data_width,
                                    size=self.banks,
                                    explicit_array=True,
                                    packed=True)
        self._write_buff_addr = self.var("write_buff_addr",
                                         self.mem_addr_width,
                                         size=self.banks,
                                         explicit_array=True,
                                         packed=True)

        self._data_out_sel = self.var("data_out_sel",
                                      self.data_width,
                                      size=self.banks,
                                      explicit_array=True,
                                      packed=True)
        self._data_addr = self.var("data_addr",
                                   self.mem_addr_width,
                                   size=self.banks,
                                   explicit_array=True,
                                   packed=True)

        self._next_num_words_mem = self.var("next_num_words_mem", 16)
        self._init_stage = self.var("init_stage", 1)
        self._read_to_write = self.var("read_to_write", 1)
        self._passthru = self.var("passthru", 1)

        self._passthru_reg = self.var("passthru_reg", self.data_width)
        self._wen_mem_en = self.var("wen_mem_en", self.banks)
        self._data_out_reg = self.var("data_out_reg", self.data_width)
        self._ren_mem_reg = self.var("ren_mem_reg", self.banks)
        self._same_bank = self.var("same_bank", self.banks)

        self._empty_d1 = self.var("empty_d1", 1)
        self._write_d1 = self.var("write_d1", 1)
        # LOCAL VARIABLES: end

        # GENERATION LOGIC: begin
        self.wire(self._read_addr_mem,
                  self._read_addr[self.full_addr - 1, self.bank_addr_width])
        self.wire(self._write_addr_mem,
                  self._write_addr[self.full_addr - 1, self.bank_addr_width])
        self.wire(self._almost_count_extended,
                  concat(const(0, 12), self._almost_count))
        self.wire(self._almost_empty,
                  self._num_words_mem <= self._almost_count_extended)
        self.wire(self._almost_full,
                  self._num_words_mem >= (self._depth - self._almost_count_extended))

        self.wire(self._empty, self._num_words_mem == 0)
        self.wire(self._full, self._num_words_mem == self._depth)

        self.wire(self._same_bank, self._ren_mem & self._wen_mem)
        self.wire(self._cen_mem, self._ren_mem | self._wen_mem_en)

        self.wire(self._ren_mem[0],
                  self._ren &
                  ((self._read_addr[self.bank_addr_width - 1, 0] == 0) | self._init_stage))
        for i in range(self.banks - 1):
            self.wire(self._ren_mem[i + 1],
                      self._ren &
                      (self._read_addr[self.bank_addr_width - 1, 0] == (i + 1)))

        self.wire(self._wen_mem[0],
                  self._wen &
                  ((self._write_addr[self.bank_addr_width - 1, 0] == 0) | self._init_stage))
        for i in range(self.banks - 1):
            self.wire(self._wen_mem[i + 1],
                      self._wen &
                      (self._write_addr[self.bank_addr_width - 1, 0] == (i + 1)))

        for i in range(self.banks):
            self.wire(self._wen_mem_en[i],
                      (self._wen_mem[i] & ~self._same_bank[i]) | self._write_buffed[i])

        for i in range(self.banks):
            self.wire(self._fifo_to_mem_data[i],
                      ternary(self._write_buffed[i],
                              self._write_buff[i],
                              self._data_in))

        self.wire(self._fifo_to_mem_cen, self._cen_mem)
        self.wire(self._fifo_to_mem_wen, self._wen_mem_en)
        self.wire(self._fifo_to_mem_addr, self._data_addr)
        self.wire(self._data_out_sel, self._mem_to_fifo_data)

        self.data_out_comb = self.combinational()
        self.data_out_if = self.data_out_comb.if_(self._ren_mem_reg[0] &
                                                  (~self._empty_d1 | self._write_d1))
        self.data_out_if.then_(self._data_out.assign(ternary(self._passthru,
                                                             self._passthru_reg,
                                                             self._data_out_sel[0])))
        for i in range(self.banks - 1):
            self.temp_if = IfStmt(self._ren_mem_reg[i + 1] & (~self._empty_d1 | self._write_d1))
            self.temp_if.then_(self._data_out.assign(ternary(self._passthru,
                                                             self._passthru_reg,
                                                             self._data_out_sel[i + 1])))
            self.data_out_if.else_(self.temp_if)
            self.data_out_if = self.temp_if
        self.data_out_if.else_(self._data_out.assign(self._data_out_reg))

        self.add_code(self.empty_d1_reg)
        self.add_code(self.write_d1_reg)
        self.add_code(self.valid_update)
        self.add_code(self.num_words_update)
        self.add_code(self.next_num_words_comb)
        self.add_code(self.data_addr_data_out_comb)
        self.add_code(self.read_addr_update)
        self.add_code(self.write_addr_update)
        self.add_code(self.init_stage_update)
        self.add_code(self.read_to_write_stage_update)
        self.add_code(self.data_out_reg_update)
        self.add_code(self.passthru_update)
        self.add_code(self.passthru_reg_update)
        self.add_code(self.write_buffed_update)
        self.add_code(self.write_buff_update)
        self.add_code(self.write_buff_addr_update)
        self.add_code(self.ren_mem_reg_update)

        # GENERATION LOGIC: end

    @always((posedge, "clk"), (posedge, "reset"))
    def empty_d1_reg(self):
        if self._reset:
            self._empty_d1 = 0
        elif self._clk_en:
            self._empty_d1 = ~self._flush & self._empty

    @always((posedge, "clk"), (posedge, "reset"))
    def write_d1_reg(self):
        if self._reset:
            self._write_d1 = 0
        elif self._clk_en:
            self._write_d1 = ~self._flush & self._wen

    @always((posedge, "clk"), (posedge, "reset"))
    def valid_update(self):
        if self._reset:
            self._valid = 0
        elif self._clk_en:
            self._valid = ~self._flush & (self._ren & (~self._empty | self._wen))

    @always((posedge, "clk"), (posedge, "reset"))
    def num_words_update(self):
        if self._reset:
            self._num_words_mem = 0
        elif self._clk_en:
            if self._flush:
                self._num_words_mem = 0
            else:
                self._num_words_mem = self._num_words_mem + self._next_num_words_mem

    def next_num_words_comb(self):
        if self._ren & self._wen:
            self._next_num_words_mem = 0
        elif self._ren & ~self._empty:
            self._next_num_words_mem = 65535
        elif self._wen & ~self._full:
            self._next_num_words_mem = 1
        else:
            self._next_num_words_mem = 0

    def data_addr_data_out_comb(self):
        for i in range(self.banks):
            self._data_addr[i] = ternary(self._write_buffed[i],
                                         self._write_buff_addr[i],
                                         ternary(self._ren_mem[i],
                                                 self._read_addr_mem,
                                                 self._write_addr_mem))

    @always((posedge, "clk"), (posedge, "reset"))
    def read_addr_update(self):
        if self._reset:
            self._read_addr = 0
        elif self._clk_en:
            if self._flush:
                self._read_addr = 0
            # READ AND NO WRITE
            elif self._ren & ~self._wen:
                if self._circular_en & ~self._empty:
                    if (self._read_addr + 1) == self._write_addr:
                        self._read_addr = 0
                    else:
                        self._read_addr = self._read_addr + 1
                elif ~self._empty:
                    self._read_addr = self._read_addr + 1
            # READ AND WRITE
            elif self._ren & self._wen:
                self._read_addr = self._read_addr + 1

    @always((posedge, "clk"), (posedge, "reset"))
    def write_addr_update(self):
        if self._reset:
            self._write_addr = 0
        elif self._clk_en:
            if self._flush:
                self._write_addr = 0
            # WRITE AND NO READ AND NOT FULL
            elif self._wen & ~self._ren & ~self._full:
                self._write_addr = self._write_addr + 1
            # READ AND WRITE
            elif self._ren & self._wen:
                self._write_addr = self._write_addr + 1

    @always((posedge, "clk"), (posedge, "reset"))
    def init_stage_update(self):
        if self._reset:
            self._init_stage = 1
        elif self._clk_en:
            if self._flush:
                self._init_stage = 1
            # Transition out of the init stage after a read or write
            elif self._wen | self._ren:
                self._init_stage = 0

    @always((posedge, "clk"), (posedge, "reset"))
    def read_to_write_stage_update(self):
        if self._reset:
            self._read_to_write = 1
        elif self._clk_en:
            if self._flush:
                self._read_to_write = 1
            # Transition out of the init stage after a read or write
            # READ AND NO WRITE
            elif self._ren & ~self._wen:
                if self._circular_en & ~self._empty:
                    if (self._read_addr + 1) == self._write_addr:
                        self._read_to_write = 1
                    else:
                        self._read_to_write = 0
                elif ~self._empty:
                    if (self._read_addr + 1) == self._write_addr:
                        self._read_to_write = 1
                    else:
                        self._read_to_write = 0
            # WRITE AND NO READ
            elif self._wen & ~self._ren & ~self._full:
                self._read_to_write = 0

    @always((posedge, "clk"), (posedge, "reset"))
    def data_out_reg_update(self):
        if self._reset:
            self._data_out_reg = 0
        elif self._clk_en:
            if self._flush:
                self._data_out_reg = 0
            # Transition out of the init stage after a read or write
            else:
                self._data_out_reg = self._data_out

    @always((posedge, "clk"), (posedge, "reset"))
    def passthru_update(self):
        if self._reset:
            self._passthru = 0
        elif self._clk_en:
            if self._flush:
                self._passthru = 0
            # Transition out of the init stage after a read or write
            elif self._ren & ~self._wen & ~self._empty:
                self._passthru = 0
            elif self._wen & ~self._ren & ~self._full:
                self._passthru = 0
            elif self._ren & self._wen:
                if self._empty & (self._same_bank.r_or()):
                    self._passthru = 1
                else:
                    self._passthru = 0

    @always((posedge, "clk"), (posedge, "reset"))
    def passthru_reg_update(self):
        if self._reset:
            self._passthru_reg = 0
        elif self._clk_en:
            if self._flush:
                self._passthru_reg = 0
            # Transition out of the init stage after a read or write
            elif self._ren & self._wen:
                if self._empty & self._same_bank.r_or():
                    self._passthru_reg = self._data_in

    @always((posedge, "clk"), (posedge, "reset"))
    def write_buffed_update(self):
        if self._reset:
            self._write_buffed = 0
        elif self._clk_en:
            if self._flush:
                self._write_buffed = 0
            else:
                # We know that it's 1 hot
                for i in range(self.banks):
                    if self._same_bank[i]:
                        self._write_buffed[i] = 1
                    elif self._write_buffed[i]:
                        self._write_buffed[i] = 0

    @always((posedge, "clk"), (posedge, "reset"))
    def write_buff_update(self):
        if self._reset:
            self._write_buff = 0
        elif self._clk_en:
            if self._flush:
                self._write_buff = 0
            else:
                # We know that it's 1 hot
                for i in range(self.banks):
                    if self._same_bank[i]:
                        self._write_buff[i] = self._data_in

    @always((posedge, "clk"), (posedge, "reset"))
    def write_buff_addr_update(self):
        if self._reset:
            self._write_buff_addr = 0
        elif self._clk_en:
            if self._flush:
                self._write_buff_addr = 0
            else:
                # We know that it's 1 hot
                for i in range(self.banks):
                    if self._same_bank[i]:
                        self._write_buff_addr[i] = self._write_addr_mem

    @always((posedge, "clk"), (posedge, "reset"))
    def ren_mem_reg_update(self):
        if self._reset:
            self._ren_mem_reg = 0
        elif self._clk_en:
            if self._flush:
                self._ren_mem_reg = 0
            # Transition out of the init stage after a read or write
            else:
                self._ren_mem_reg = self._ren_mem


if __name__ == "__main__":
    fifo_dut = FIFOControl(16, 2, 512)
    verilog(fifo_dut, filename="fifo_control.sv", check_active_high=False)



import kratos as kts
from enum import Enum
from enum import auto


class FormalSignalConstraint(Enum):
    X = auto()
    SET0 = auto()
    SET1 = auto()
    CLK = auto()
    RSTN = auto()
    SOLVE = auto()
    SEQUENCE = auto()


class FormalAttr(kts.Attribute):
    def __init__(self,
                 port_name,
                 formalsig_cnst,
                 module="all",
                 doc_string=""):
        super().__init__()
        self.port_name = port_name
        self.formalsig_cnst = formalsig_cnst
        self.module = module
        self.documentation = doc_string

    def set_documentation(self, new_doc):
        self.documentation = new_doc

    def get_documentation(self):
        return self.documentation

    def get_annotation(self):
        # need annotation to be updated if port_name is updated
        # after creation like during cut_generator pass
        return f"{self.port_name}\t{self.formalsig_cnst.name}"

    def get_module(self):
        return self.module

    def get_port_name(self):
        return self.port_name

    def get_formal_ann(self):
        return self.formalsig_cnst



import kratos as kts
import _kratos
from lake.attributes.config_reg_attr import ConfigRegAttr
from lake.attributes.control_signal_attr import ControlSignalAttr
from lake.passes.passes import lift_config_reg, change_sram_port_names
from lake.top.lake_top import LakeTop
from lake.utils.sram_macro import SRAMMacroInfo
import collections


CFG_info = collections.namedtuple('CFG_info', 'port_name port_size port_width expl_arr')
IO_info = collections.namedtuple('IO_info', 'port_name port_size port_width is_ctrl port_dir expl_arr')


def extract_top_config(circuit_gen: kts.Generator):
    int_gen = circuit_gen.internal_generator
    config_list = []

    # Now get the config registers from the top definition
    for port_name in int_gen.get_port_names():
        curr_port = int_gen.get_port(port_name)
        attrs = curr_port.find_attribute(lambda a: isinstance(a, ConfigRegAttr))
        if len(attrs) != 1:
            continue
        cr_attr = attrs[0]
        config_list.append(CFG_info(port_name=port_name,
                                    port_size=curr_port.size,
                                    port_width=curr_port.width,
                                    expl_arr=curr_port.explicit_array))
    return config_list


def get_interface(circuit_gen: kts.Generator):
    intf_sigs = []
    int_gen = circuit_gen.internal_generator
    # Now get the config registers from the top definition
    for port_name in int_gen.get_port_names():
        curr_port = int_gen.get_port(port_name)
        attrs = curr_port.find_attribute(lambda a: isinstance(a, ControlSignalAttr))
        if len(attrs) != 1:
            continue
        cr_attr = attrs[0]
        # Now we have this
        intf_sigs.append(IO_info(port_name=port_name,
                                 port_size=curr_port.size,
                                 port_width=curr_port.width,
                                 is_ctrl=cr_attr.get_control(),
                                 port_dir=str(curr_port.port_direction),
                                 expl_arr=curr_port.explicit_array))
    return intf_sigs


if __name__ == "__main__":
    tsmc_info = SRAMMacroInfo("tsmc_name")
    use_sram_stub = False
    fifo_mode = True
    mem_width = 64
    lake_dut = LakeTop(mem_width=mem_width,
                       sram_macro_info=tsmc_info,
                       use_sram_stub=use_sram_stub,
                       fifo_mode=fifo_mode,
                       add_clk_enable=True,
                       add_flush=True)
    sram_port_pass = change_sram_port_names(use_sram_stub=use_sram_stub, sram_macro_info=tsmc_info)
    # Perform pass to move config_reg
    extract_top_config(lake_dut)
    # get_interface(lake_dut)



from lake.models.model import Model
import kratos as kts


class SyncGroupsModel(Model):
    def __init__(self,
                 fetch_width,
                 data_width,
                 int_out_ports):

        self.fetch_width = fetch_width
        self.data_width = data_width
        self.fw_int = int(self.fetch_width / self.data_width)
        self.int_out_ports = int_out_ports
        self.groups = self.int_out_ports

        self.config = {}
        for i in range(self.int_out_ports):
            self.config[f'sync_group_{i}'] = 0

        self.local_gate = []
        self.local_mask = []

        for i in range(self.groups):
            self.local_gate.append([])
            self.local_mask.append([])
            for j in range(self.int_out_ports):
                self.local_gate[i].append(1)
                self.local_mask[i].append(1)

        self.local_gate_reduced = []
        self.sync_group_valid = []
        self.valid_reg = []
        self.data_reg = []
        self.mem_valid_data_out_reg = []
        for i in range(self.int_out_ports):
            self.local_gate_reduced.append(1)
            self.sync_group_valid.append(0)
            self.valid_reg.append(0)
            self.mem_valid_data_out_reg.append(0)
            row = []
            for j in range(self.fetch_width):
                row.append(0)
            self.data_reg.append(row)

    def set_config(self, new_config):
        # No configuration space
        # Configure top level
        for key, config_val in new_config.items():
            if key not in self.config:
                AssertionError("Gave bad config...")
            else:
                self.config[key] = config_val

    def interact(self,
                 ack_in,
                 data_in,
                 valid_in,
                 ren_in,
                 mem_valid_data):
        '''
        Returns (data_out, valid_out, rd_sync_gate, mem_valid_data_out)
        '''
        valid_out = []
        data_out = []
        rd_sync_gate = []
        mem_valid_data_out = []
        # # Use current state of bus to set local gate reduced
        # for i in range(self.int_out_ports):
        #     # For this port, just want to check that its corresponding entry is low
        #     for j in range(self.groups):
        #         if self.config[f"sync_group_{i}"] == (1 << j):
        #             self.local_gate_reduced[i] = self.local_gate[j][i]

        # Set the ren_int, ack_in combo
        ren_int = []
        for i in range(self.int_out_ports):
            ren_int.append(ren_in[i] & self.local_gate_reduced[i])

        # Create new local mask
        for i in range(self.groups):
            for j in range(self.int_out_ports):
                self.local_mask[i][j] = 1
                # If port j is in group i, set its gate mask
                if self.config[f"sync_group_{j}"] == (1 << i):
                    self.local_mask[i][j] = not (ren_int[j] & ((ack_in & (1 << j)) != 0))
                    # self.local_mask[i][j] = not (ren_int[j] and ack_in[j])

        # Get group finished
        group_finished = []
        for i in range(self.groups):
            group_finished.append(1)
            # Check that either the bus or mask is low for all items in the group
            for j in range(self.int_out_ports):
                # Only check if the port is in the group
                if self.config[f"sync_group_{j}"] == (1 << i):
                    if self.local_gate[i][j] == 1 and self.local_mask[i][j] == 1:
                        group_finished[i] = 0

        rd_sync_gate = self.get_rd_sync()

        for i in range(self.groups):
            for j in range(self.int_out_ports):
                if group_finished[i] == 1:
                    self.local_gate[i][j] = 1
                else:
                    self.local_gate[i][j] = self.local_gate[i][j] & self.local_mask[i][j]

        # Can get the valid syncs now
        # We do this by checking each groups members
        # to all have their valid reg high
        for i in range(self.groups):
            self.sync_group_valid[i] = 1
            for j in range(self.int_out_ports):
                # If any member of the group isn't valid yet, the group isn't valid
                if (self.config[f"sync_group_{j}"] == (1 << i)) and self.valid_reg[j] == 0:
                    self.sync_group_valid[i] = 0

        # Each port gets its group's sync valid
        for i in range(self.int_out_ports):
            valid_out.append(self.sync_group_valid[kts.clog2(self.config[f'sync_group_{i}'])])
            # valid_out.append(self.sync_group_valid[self.config[f'sync_group_{i}']])
            data_out.append(self.data_reg[i].copy())
            mem_valid_data_out.append(self.mem_valid_data_out_reg[i])

        # Update the registered valids - we keep these around to track
        # which valids in the group already came
        for i in range(self.int_out_ports):
            group_log = kts.clog2(self.config[f"sync_group_{i}"])
            if self.sync_group_valid[group_log] == 1 or self.valid_reg[i] == 0:
                self.valid_reg[i] = valid_in[i]
                self.data_reg[i] = data_in[i].copy()
                self.mem_valid_data_out_reg[i] = mem_valid_data[i]

        # Use current state of bus to set local gate reduced
        for i in range(self.int_out_ports):
            # For this port, just want to check that its corresponding entry is low
            for j in range(self.groups):
                if self.config[f"sync_group_{i}"] == (1 << j):
                    self.local_gate_reduced[i] = self.local_gate[j][i]

        # rd_sync_gate = self.get_rd_sync()
        return (data_out.copy(), valid_out, rd_sync_gate.copy(), mem_valid_data_out)

    def get_rd_sync(self):
        rd_sync_gate = []
        for i in range(self.int_out_ports):
            rd_sync_gate.append(self.local_gate_reduced[i])
        return list(rd_sync_gate)
